class MLError(Exception):
    """
    Exception class from which every exception in this library will derive.
    It enables other projects using this library to catch all errors coming
    from the library with a single "except" statement
    """

    pass


class CatalogError(MLError):
    """
    Raised when the catalog is not valid
    """

    pass


class InvalidDataFormatError(MLError):
    """
    Raised when the catalog is not valid
    """

    pass


class LocalDirNotWriteableException(MLError):
    """
    Raised when the local directory is not writeable
    """

    pass


class MissingConfigFileException(MLError):
    """
    Raised when a given configuration file cannot be found
    """

    pass


class BadCLIParameterException(MLError):
    """
    Raised when there is an issue with a parameter
    in the CLI (commnand-line interface)
    """

    pass


class BadConfigException(MLError):
    """
    Raised when a configuration file cannot be loaded, for instance
    due to wrong syntax or poor formatting.
    """

    pass


class BadConfigLogLevelException(BadConfigException):
    """
    Raised when the log level (`log_level`) does not exist
    """

    pass


class BadConfigSparkMasterException(BadConfigException):
    """
    Raised when the Spark master (`spark_master`) is not valid
    """

    pass


class BadConfigPathException(BadConfigException):
    """
    Raised when the configuration path parameter (`conf_path`) does not
    contain at least a directory/path where to find configuration files
    """

    pass


class BadConfigMissingInputException(BadConfigException):
    """
    Raised when the 'input_data'/{'products', 'transactions'} file-path
    is missing
    """

    pass


class BadConfigMissingOutputException(BadConfigException):
    """
    Raised when the 'output_data'/'transactions' file-path is missing
    """

    pass


class TaskNotFoundError(MLError):
    """Raised when task name is not found in entrypoints"""


class MissingDatasetError(MLError):
    """Raised when a dataset is not found in the catalog"""

    pass


class DataSetError(MLError):
    """Raised when there is an issue with a dataset"""

    pass


class MissingConfigException(MLError):
    """Raised when a configuration is missing"""

    pass


class BadConfigFormatException(MLError):
    """Raised when a configuration is not formatted correctly"""

    pass


class InconsistentDatasetConfigurations(MLError):
    """
    Raised when several inconsistent dataset configurations are being used
    to build a dataset.
    """

    pass


class NotMonotone(MLError):
    """
    Raised when some outputs are not a monotone function of some inputs.
    """

    pass


class NotNonDecreasing(MLError):
    """
    Raised when some outputs are not an non-decreasing function of some inputs.
    """

    pass


class NotNonIncreasing(MLError):
    """
    Raised when some outputs are not an non-increasing function of some inputs.
    """

    pass



from __future__ import annotations

from collections import OrderedDict, UserList
from dataclasses import dataclass
from typing import Any, List, Optional

import numpy as np
import pandas as pd


@dataclass(frozen=True)
class DataFrameRow:
    """
    A single row of a pandas dataframe.
    The motivation for this abstraction is to speed up concatenation
    of a list of rows (see function `concat_df_rows`).
    Only integer indices are supported
    """

    values: OrderedDict[str, Any]
    dtypes: OrderedDict[str, np.dtype]
    index: int

    def __post_init__(self):
        if not self.values.keys() == self.dtypes.keys():
            raise ValueError("values and dtypes should have the same set of keys")

    @classmethod
    def from_single_row_df(cls, single_row_df: pd.DataFrame) -> DataFrameRow:
        if len(single_row_df) != 1:
            raise ValueError("single_row_df should contain exactly one row")
        return cls(
            values=single_row_df.iloc[0].to_dict(OrderedDict),
            dtypes=single_row_df.dtypes.to_dict(OrderedDict),
            index=single_row_df.index[0],
        )

    @property
    def columns(self) -> List[str]:
        return list(self.dtypes.keys())

    @property
    def single_row_df(self) -> pd.DataFrame:
        return pd.DataFrame(
            {
                column: np.array([self.values[column]], dtype=self.dtypes[column])
                for column in self.columns
            },
            index=[self.index],
        )

    def __getitem__(self, column: str) -> Any:
        return self.values[column]

    def __contains__(self, column: str) -> bool:
        return column in self.columns

    def set(self, column: str, value: Any, dtype: Optional[np.dtype] = None) -> DataFrameRow:
        new_values = OrderedDict(self.values)
        new_values.update({column: value})
        new_dtypes = OrderedDict(self.dtypes)
        if column not in self.columns or dtype:
            new_dtypes.update({column: np.array(value, dtype=dtype).dtype})
        return DataFrameRow(
            new_values,
            new_dtypes,
            self.index,
        )

    def __repr__(self) -> str:
        return (
            "{\n"
            + "\n".join(
                [
                    f"   {column} ({self.dtypes[column]}): {self.values[column]}"
                    for column in self.columns
                ]
            )
            + "\n}"
        )


class ListDataFrameRows(UserList[DataFrameRow]):
    """
    This class improves the speed of execution of some
    of the built-in operations of pandas dataframes like
    `__getitem__` (subscript),
    """

    def __init__(
        self,
        columns: List[str],
        dtypes: OrderedDict[str, np.dtype],
        data: List[DataFrameRow],
    ):
        """
        Avoid using the constructor directly, prefer `from_pandas_df`
        and `from_list_rows` methods.
        """
        self.columns = columns
        self.dtypes = dtypes
        self.data = data

    @classmethod
    def from_pandas_df(
        cls,
        df: pd.DataFrame,
    ):
        columns = df.columns
        dtypes = df.dtypes.to_dict(OrderedDict)
        column_values = {column: df[column].to_numpy() for column in df.columns}
        indices = df.index
        data: List[DataFrameRow] = [
            DataFrameRow(
                OrderedDict({column: column_values[column][i] for column in df.columns}),
                dtypes,
                indices[i],
            )
            for i in range(len(df))
        ]
        return cls(columns=columns, dtypes=dtypes, data=data)

    @classmethod
    def from_list_rows(cls, list_rows: List[DataFrameRow]):
        """
        Warning: not checking rows consistency to avoid slowing down execution.
        Check rows consistency before calling this method if needed.
        """
        return cls(columns=list_rows[0].columns, dtypes=list_rows[0].dtypes, data=list_rows)

    def get_column(self, column: str) -> np.ndarray:
        return np.array([row[column] for row in self], dtype=self.dtypes[column])

    def build_dataframe(self, columns: List[str]) -> pd.DataFrame:
        """
        Convert list of rows back to a pandas dataframe (only keeping
        columns in `columns`).
        Concatenating a list of `DataFrameRow` is much faster than
        concatenating a list of single-row `pandas.DataFrame`
        using `pandas.concat` function.
        """
        return pd.DataFrame(
            {column: self.get_column(column) for column in self.columns if column in columns},
            index=[row.index for row in self],
        )

    @property
    def indices(self):
        return [row.index for row in self]

    @property
    def dataframe(
        self,
    ) -> pd.DataFrame:
        return self.build_dataframe(self.columns)



from __future__ import annotations

import itertools
from dataclasses import dataclass
from inspect import signature
from typing import Any, Dict, Generic, List, Optional, Tuple, Type, TypeVar

import numpy as np
import pandas as pd
import plotly
import plotly.express as px
import pytorch_lightning as pl
import torch
import torch.nn as nn

from pytorch_utils.data_modules import (
    AugmentedBernoulliDataModule,
)
from pytorch_utils.utils import (
    BatchTorchTensors,
    CategoricalFeatureEmbedding,
    NamedTorchMetrics,
    NamedTorchTensors,
    assert_monotone,
)


class LinearNonNeg(nn.Linear):
    """
    Alternative linear layer with nonnegative weights (bias unchanged).
    This ensures the outputs are always a non-decreasing function of the inputs
    (no matter the values of parameters `self.weight` and `self.bias`, which may vary during training).

    The easiest way to implement this class with minimal code is to subclass `torch.nn.Linear` and
    apply a positive transformation (namely `torch.nn.functional.elu` shifted by 1) to the weights
    before applying the linear transformation in the `forward` method.
    """

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return nn.functional.linear(
            input,
            torch.nn.functional.elu(  # this is the only difference with the original torch.nn.Linear module
                self.weight, alpha=1.0, inplace=False
            )
            + 1,
            self.bias,
        )


class BatchNorm1dNonNeg(nn.BatchNorm1d):
    """
    Alternative batch normalization with nonnegative weights (bias unchanged).
    This ensures the outputs are always a non-decreasing function of the inputs
    when `self.training=False` (no matter the values of parameters `self.weight` and `self.bias`,
    which may vary during training).

    The easiest way to implement this class with minimal code is to subclass
    `torch.nn.BatchNorm1d` and apply a positive transformation (namely `torch.nn.functional.elu` shifted by 1)
    to the weights before applying the batch norm transformation in the `forward` method.
    """

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        self._check_input_dim(input)

        # exponential_average_factor is set to self.momentum
        # (when it is available) only so that it gets updated
        # in ONNX graph when this node is exported to ONNX.
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum

        if self.training and self.track_running_stats:
            # TODO: if statement only here to tell the jit to skip emitting this when it is None
            if self.num_batches_tracked is not None:  # type: ignore[has-type]
                self.num_batches_tracked.add_(1)  # type: ignore[has-type]
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum

        r"""
        Decide whether the mini-batch stats should be used for normalization rather than the buffers.
        Mini-batch stats are used in training mode, and in eval mode when buffers are None.
        """
        if self.training:
            bn_training = True
        else:
            bn_training = (self.running_mean is None) and (self.running_var is None)

        r"""
        Buffers are only updated if they are to be tracked and we are in training mode. Thus they only need to be
        passed when the update should occur (i.e. in training mode when they are tracked), or when buffer stats are
        used for normalization (i.e. in eval mode when buffers are not None).
        """

        return torch.nn.functional.batch_norm(
            input,
            # If buffers are not to be tracked, ensure that they won't be updated
            self.running_mean if not self.training or self.track_running_stats else None,
            self.running_var if not self.training or self.track_running_stats else None,
            torch.nn.functional.elu(  # this is the only difference with the original torch.nn.Linear module
                self.weight, alpha=1.0, inplace=False
            )
            + 1,
            self.bias,
            bn_training,
            exponential_average_factor,
            self.eps,
        )


class LinearSemiNonNeg(nn.Module):
    """
    Alternative linear layer combining a standard linear layer (`torch.nn.Linear`)
    together with a `LinearNonNeg` layer (by summing the two).
    The outputs are always a non-decreasing function of the inputs named `non_neg_inputs_name`
    (no matter the weights and biases).
    The outputs are not necessarily monotone w.r.t. the inputs named `other_inputs_name`.
    """

    def __init__(
        self,
        in_features_non_neg,
        in_features_others,
        out_features,
        non_neg_inputs_name="non_neg_inputs",
        other_inputs_name="other_inputs",
    ):
        super().__init__()
        self.in_features_non_neg = in_features_non_neg
        self.in_features_others = in_features_others
        self.out_features = out_features
        self.non_neg_inputs_name = non_neg_inputs_name
        self.other_inputs_name = other_inputs_name

        self.linear_non_neg = LinearNonNeg(
            in_features=self.in_features_non_neg,
            out_features=self.out_features,
            bias=False,  # duplicating biases is useless
        )
        self.linear_others = nn.Linear(
            in_features=self.in_features_others,
            out_features=self.out_features,
            bias=True,
        )

    def forward(self, input: NamedTorchTensors):
        return self.linear_non_neg(input[self.non_neg_inputs_name]) + self.linear_others(
            input[self.other_inputs_name]
        )


class BiLinearSemiNonNeg(nn.Module):
    def __init__(
        self,
        in_features_non_neg,
        in_features_others,
        out_features_non_neg,
        out_features_others,
        non_neg_inputs_name="non_neg_inputs",
        other_inputs_name="other_inputs",
    ):
        """
        Yet another custom layer that concatenates a standard
        linear layer (`torch.nn.Linear`) together with a `LinearNonNeg`
        layer (keeping the two layers separate).
        """
        super().__init__()
        self.in_features_non_neg = in_features_non_neg
        self.in_features_others = in_features_others
        self.out_features_non_neg = out_features_non_neg
        self.out_features_others = out_features_others
        self.non_neg_inputs_name = non_neg_inputs_name
        self.other_inputs_name = other_inputs_name

        self.linear_semi_non_neg = LinearSemiNonNeg(
            in_features_non_neg=self.in_features_non_neg,
            in_features_others=self.in_features_others,
            out_features=self.out_features_non_neg,
            non_neg_inputs_name=self.non_neg_inputs_name,
            other_inputs_name=self.other_inputs_name,
        )
        self.linear = nn.Linear(
            in_features=self.in_features_others,
            out_features=self.out_features_others,
            bias=True,
        )

    def forward(self, input: NamedTorchTensors):
        return {
            self.other_inputs_name: self.linear(input[self.other_inputs_name]),
            self.non_neg_inputs_name: self.linear_semi_non_neg(input),
        }


class Partitioned(nn.Module):
    """
    Unlike `torch.nn.Sequential` wich “chains” outputs to inputs
    sequentially for each module in a provided list, this module
    simultaneously transforms every partition of the input in parallel
    using the corresponding module.
    The difference between `torch.nn.Sequential` and `Partitioned` is
    similar to the difference between a series and parallel electric circuit.
    """

    def __init__(self, **module_partitions: nn.Module):
        super().__init__()
        self.modules_dict = nn.ModuleDict(module_partitions)

    def forward(self, input_partitions: NamedTorchTensors):
        return {k: module(input_partitions[k]) for k, module in self.modules_dict.items()}


class ShiftedEmbedding(nn.Embedding):
    """
    Custom embedding module that shifts all indices by 1.
    The original `torch.nn.Embedding` layer only accepts non-negative integers as inputs.
    This custom layer accepts non-negative integers and -1 as inputs.
    This is useful when -1 is used to encode unknown and/or missing values
    (i.e., using a sklearn.preprocessing.OrdinalEncoder with unknown_value=-1 and/or encoded_missing_value=-1).

    The easiest way to implement this class with minimal code is to subclass
    `torch.nn.Embedding` and shift the inputs by 1 in `forward` method.
    """

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return nn.functional.embedding(
            1 + input,  # this is the only difference with the original torch.nn.Embedding module
            self.weight,
            self.padding_idx,
            self.max_norm,
            self.norm_type,
            self.scale_grad_by_freq,
            self.sparse,
        )


class MeanImputationEmbedding(nn.Embedding):
    """
    Custom embedding module that applies "mean imputation" when inputs negative.

    The original `torch.nn.Embedding` layer only accepts non-negative integers as inputs.
    This custom layer also accepts negative integers as inputs.
    This is useful when for instance -1 is used to encode unknown and/or missing values
    (i.e., using a sklearn.preprocessing.OrdinalEncoder with unknown_value=-1 and/or encoded_missing_value=-1).
    When a negative input is provided, all embeddings are averaged (form of "mean imputation").
    """

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return torch.where(
            (input > -1)[..., None],
            nn.functional.embedding(
                nn.ReLU()(input),
                self.weight,
                self.padding_idx,
                self.max_norm,
                self.norm_type,
                self.scale_grad_by_freq,
                self.sparse,
            ),
            nn.functional.embedding(
                torch.tensor([0], device=input.device),
                self.weight.mean(dim=0)[None, ...],
                self.padding_idx,
                self.max_norm,
                self.norm_type,
                self.scale_grad_by_freq,
                self.sparse,
            ),
        )


class MonotoneBernoulliProbability(pl.LightningModule):
    """
    Predicts the probability of success of an event conditioned on some features.
    The structure of the neural network enforces that the predicted probability
    is a monotone (i.e., non-decreasing and/or non-increasing) function of some specified
    features.

    `optimizer_params` should at least contain the keys `class` and `lr`

    Two ways of doing inference:
        - use `self.predict` directly
        - use method `predict` of `pytorch_lightning.Trainer`
    """

    module_scope: str

    def __init__(
        self,
        real_features_non_decreasing: List[str],
        real_features_non_increasing: List[str],
        real_features_non_monotone: List[str],
        categorical_feature_embeddings: List[CategoricalFeatureEmbedding] = [],
        hidden_sizes_monotone: List[int] = [],
        hidden_sizes_non_monotone: List[int] = [],
        polynomial_real_features_expansions: Dict[str, List[int]] = dict(),
        activation_layer_monotone: Type[torch.nn.Module] = nn.ReLU,
        activation_layer_non_monotone: Type[torch.nn.Module] = nn.ReLU,
        normalization_layer_monotone: Type[torch.nn.Module] = BatchNorm1dNonNeg,
        normalization_layer_non_monotone: Type[torch.nn.Module] = nn.BatchNorm1d,
        dropout_rate_monotone: int = 0,
        dropout_rate_non_monotone: int = 0,
        optim_criterion_params: Dict[str, Any] = {
            "class": nn.BCEWithLogitsLoss,
        },
        optimizer_params: Dict[str, Any] = {
            "class": torch.optim.Adam,
            "lr": 0.001,
            "betas": (0.9, 0.999),
            "eps": 1e-08,
        },
        validation_metrics: NamedTorchMetrics = {},
        test_metrics: NamedTorchMetrics = {},
    ):
        super().__init__()

        # Check inputs
        # -------------------
        if set(real_features_non_decreasing).intersection(real_features_non_increasing):
            raise ValueError(
                "real_features_non_decreasing and real_features_non_increasing must be disjoint"
            )

        if set(real_features_non_decreasing).intersection(real_features_non_monotone):
            raise ValueError(
                "real_features_non_decreasing and real_features_non_monotone must be disjoint"
            )

        if set(real_features_non_increasing).intersection(real_features_non_monotone):
            raise ValueError(
                "real_features_non_increasing and real_features_non_monotone must be disjoint"
            )

        if not set(polynomial_real_features_expansions.keys()).issubset(
            set(real_features_non_decreasing)
            .union(real_features_non_increasing)
            .union(real_features_non_monotone)
        ):
            raise ValueError(
                """features in polynomial_real_features_expansions should belong to either
                real_features_non_decreasing or real_features_non_increasing or
                real_features_non_monotone"""
            )

        for feat, degrees in polynomial_real_features_expansions.items():
            for degree in degrees:
                if degree <= 1:
                    raise ValueError("all degrees of polynom expansions should be > 1")
                if (
                    (feat in real_features_non_decreasing) or (feat in real_features_non_increasing)
                ) and degree % 2 == 0:
                    raise ValueError(
                        """the degrees of polynom expansions should be odd integers 
                        for monotone (i.e., non-decreasing or non-increasing) features
                        so as to preserve monotonicity"""
                    )

        # Features
        # -------------------
        self.real_features_non_decreasing = real_features_non_decreasing
        self.real_features_non_increasing = real_features_non_increasing
        self.real_features_monotone = (
            self.real_features_non_decreasing + self.real_features_non_increasing
        )
        self.real_features_non_monotone = real_features_non_monotone
        self.polynomial_real_features_expansions = polynomial_real_features_expansions
        self.categorical_feature_embeddings = sorted(categorical_feature_embeddings)  # type: ignore
        self.categorical_features = [f.feature_name for f in self.categorical_feature_embeddings]
        self.size_real_features_non_decreasing = len(self.real_features_non_decreasing) + sum(
            [
                len(degrees)
                for feat, degrees in self.polynomial_real_features_expansions.items()
                if feat in self.real_features_non_decreasing
            ]
        )
        self.size_real_features_non_increasing = len(self.real_features_non_increasing) + sum(
            [
                len(degrees)
                for feat, degrees in self.polynomial_real_features_expansions.items()
                if feat in self.real_features_non_increasing
            ]
        )
        self.size_real_features_monotone = (
            self.size_real_features_non_decreasing + self.size_real_features_non_increasing
        )
        self.size_real_features_non_monotone = len(self.real_features_non_monotone) + sum(
            [
                len(degrees)
                for feat, degrees in self.polynomial_real_features_expansions.items()
                if feat in self.real_features_non_monotone
            ]
        )
        self.size_embeddings = sum(map(lambda x: x.embedding_size, categorical_feature_embeddings))
        self.size_features = (
            self.size_real_features_monotone
            + self.size_real_features_non_monotone
            + self.size_embeddings
        )
        self._monotone_feat_name = "monotone_features"
        self._non_monotone_feat_name = "non_monotone_features"

        # Neural nets
        # -------------------
        self.hidden_sizes_monotone = hidden_sizes_monotone
        self.hidden_sizes_non_monotone = hidden_sizes_non_monotone

        self.activation_layer_monotone = activation_layer_monotone
        self.activation_layer_non_monotone = activation_layer_non_monotone
        self.normalization_layer_monotone = normalization_layer_monotone
        self.normalization_layer_non_monotone = normalization_layer_non_monotone
        self.dropout_rate_monotone = dropout_rate_monotone
        self.dropout_rate_non_monotone = dropout_rate_non_monotone
        self._build_neural_nets()

        # Optimization
        # -------------------
        self.optim_criterion_params = optim_criterion_params
        self._build_optim_criterion()
        self.optimizer_params = optimizer_params

        # Metrics
        # -------------------
        # Cloning the metrics is safer: modular metrics contain internal states that
        # should belong to only one DataLoader, it is recommended to initialize a separate modular metric instances
        # for each DataLoader and in particular use separate metrics for training, validation and testing
        # see: https://torchmetrics.readthedocs.io/en/latest/pages/lightning.html?highlight=
        # modular%20metrics%20contain%20internal%20states#common-pitfalls
        self.validation_metrics = self._clone_metrics(validation_metrics)
        self.test_metrics = self._clone_metrics(test_metrics)

        # Save hyper-parameters
        # -------------------
        self.save_hyperparameters()  # saves all constructor params by default

    @property
    def learning_rate(self) -> float:
        return self.optimizer_params["lr"]

    @learning_rate.setter
    def learning_rate(self, learning_rate) -> None:
        """
        Updates learning rate (useful to apply `pytorch_lightning.trainer.Trainer.tune.lr_find`)
        """
        self.optimizer_params.update({"lr": learning_rate})

    def forward(self, x: NamedTorchTensors) -> torch.Tensor:
        return self._forward_from_logits(self._logits(x))

    def _forward_from_logits(self, logits: torch.Tensor) -> torch.Tensor:
        return self.final_sigmoid_layer(logits)

    def configure_optimizers(self):
        optimizer_class = self.optimizer_params.pop("class")
        try:
            optimizer = optimizer_class(self.parameters(), **self.optimizer_params)
        finally:
            self.optimizer_params["class"] = optimizer_class
        return optimizer

    @staticmethod
    def _clone_metrics(metrics: NamedTorchMetrics) -> nn.ModuleDict:
        return nn.ModuleDict(
            {metric_name: metric.clone() for metric_name, metric in metrics.items()}
        )

    def _shared_step(
        self,
        batch: BatchTorchTensors,
        batch_idx: int,
    ) -> Tuple[NamedTorchTensors, torch.Tensor, torch.Tensor, torch.Tensor, int]:
        features, targets, weights = batch
        batch_size = len(targets)
        logits = self._logits(features)
        preds = self._forward_from_logits(logits)
        loss = torch.mean(
            self.optim_criterion(logits, targets.float()) * (weights if weights is not None else 1)
        )
        return features, targets, weights, preds, loss, batch_size  # type: ignore

    def _evaluate_metrics(
        self,
        metrics: NamedTorchMetrics,
        preds: torch.Tensor,
        targets: torch.Tensor,
        weights: Optional[torch.Tensor] = None,
    ) -> None:
        for metric in metrics.values():
            if weights is None or len(signature(metric.update).parameters) < 4:
                metric(preds, targets)
            else:
                metric(preds, targets, weights)

    def training_step(self, batch: BatchTorchTensors, batch_idx: int) -> torch.Tensor:
        _, _, _, _, loss, batch_size = self._shared_step(batch, batch_idx)  # type: ignore
        self.log(
            "training_loss",
            loss,  # type: ignore
            batch_size=batch_size,  # type: ignore
            on_step=True,
            on_epoch=True,
            prog_bar=True,
        )
        return loss  # type: ignore

    def validation_step(self, batch: BatchTorchTensors, batch_idx: int) -> torch.Tensor:
        _, targets, weights, preds, loss, batch_size = self._shared_step(batch, batch_idx)  # type: ignore
        self.log(
            "validation_loss",
            loss,  # type: ignore
            batch_size=batch_size,  # type: ignore
            on_step=False,
            on_epoch=True,
            prog_bar=True,
        )
        if self.validation_metrics:
            self._evaluate_metrics(self.validation_metrics, preds, targets, weights)  # type: ignore
            self.log_dict(
                self.validation_metrics,  # type: ignore
                batch_size=batch_size,  # type: ignore
                on_step=False,
                on_epoch=True,
            )
        return loss  # type: ignore

    def test_step(self, batch: BatchTorchTensors, batch_idx: int) -> torch.Tensor:
        _, targets, weights, preds, loss, batch_size = self._shared_step(batch, batch_idx)  # type: ignore
        self.log(
            "test_loss",
            loss,  # type: ignore
            batch_size=batch_size,  # type: ignore
            on_step=False,
            on_epoch=True,
            prog_bar=True,
        )
        if self.test_metrics:
            self._evaluate_metrics(self.test_metrics, preds, targets, weights)  # type: ignore
            self.log_dict(
                self.test_metrics,  # type: ignore
                batch_size=batch_size,
                on_step=False,
                on_epoch=True,
            )
        return loss  # type: ignore

    def predict_step(
        self, batch: NamedTorchTensors, batch_idx: int, dataloader_idx: int = 0
    ) -> torch.Tensor:
        return self(batch)

    def _to_self_device(self, x: NamedTorchTensors):
        return {feat: x[feat] for feat in x}  # type: ignore

    def predict(
        self,
        features: NamedTorchTensors,
    ) -> torch.Tensor:
        training = self.training
        try:
            self.train(False)
            with torch.no_grad():
                predictions = self.predict_step(
                    self._to_self_device(features),
                    batch_idx=0,
                )
        finally:
            self.train(training)
        return predictions

    def predict_from_pandas(
        self,
        features: pd.DataFrame,
        data_module: AugmentedBernoulliDataModule,
        scaling_factors: np.ndarray = np.array([1.0]),
        min_augmented_value: float = -float("inf"),
        max_augmented_value: float = float("inf"),
    ) -> torch.Tensor:
        return self.predict(
            data_module.augment_transform_to_tensors(  # type: ignore
                features,
                scaling_factors,
                min_augmented_value,
                max_augmented_value,
            )
        )

    def _polynom_expansion(self, x: NamedTorchTensors) -> List[torch.Tensor]:
        return list(
            itertools.chain(
                *[
                    [x[feat] ** degree for degree in degrees]
                    for feat, degrees in self.polynomial_real_features_expansions.items()
                    if feat in x
                ]
            )
        )

    def _logits(self, x: NamedTorchTensors) -> torch.Tensor:
        # Encode categorical features into embeddings
        # -------------------
        categorical_embeddings = (
            [torch.cat([t for t in self.categorical_embeddings_net(x).values()], dim=1)]
            if len(self.categorical_feature_embeddings) > 0
            else []
        )

        # Polynom expansions of real features
        # -------------------
        non_monotone_polynom_expansions = self._polynom_expansion(
            {feat: x[feat] for feat in x if feat in self.real_features_non_monotone}
        )
        monotone_polynom_expansions = self._polynom_expansion(
            {
                feat: (x[feat] if feat in self.real_features_non_decreasing else torch.neg(x[feat]))
                for feat in x
                if feat in self.real_features_monotone
            }
        )

        # Apply core layers
        # -------------------
        non_monotone_features = torch.cat(
            categorical_embeddings
            + [x[feat] for feat in self.real_features_non_monotone]
            + non_monotone_polynom_expansions,
            dim=1,
        )
        monotone_features = torch.cat(
            [x[feat] for feat in self.real_features_non_decreasing]
            + [torch.neg(x[feat]) for feat in self.real_features_non_increasing]
            + monotone_polynom_expansions,
            dim=1,
        )

        intermediate_outputs = self.core_sequential_layers(
            {
                self._monotone_feat_name: monotone_features,
                self._non_monotone_feat_name: non_monotone_features,
            }
        )

        # Apply missing layers
        # -------------------
        logits = self.final_linear_layer(
            {
                self._monotone_feat_name: self.missing_sequential_layers_monotone(
                    intermediate_outputs[self._monotone_feat_name]
                ),
                self._non_monotone_feat_name: self.missing_sequential_layers_non_monotone(
                    intermediate_outputs[self._non_monotone_feat_name]
                ),
            }
        )

        return logits.view(-1)

    def _build_optim_criterion(self) -> None:
        optim_criterion_class = self.optim_criterion_params.pop("class")
        try:
            self.optim_criterion = optim_criterion_class(**self.optim_criterion_params)
        finally:
            self.optim_criterion_params["class"] = optim_criterion_class

    def _build_neural_nets(self) -> None:
        # Embeddings for categorical features
        # -------------------
        self._build_categorical_embeddings()

        # Core sequential layers
        # -------------------
        self._build_core_sequential_layers()

        # Missing layers
        # -------------------
        self._build_missing_layers()

        # Final layer
        # -------------------
        self._build_final_layer()

    def _build_categorical_embeddings(self) -> None:
        self.categorical_embeddings_net = Partitioned(
            **{
                cat_feat_emb.feature_name: MeanImputationEmbedding(  # ShiftedEmbedding(
                    num_embeddings=cat_feat_emb.nb_distinct_values,
                    embedding_dim=cat_feat_emb.embedding_size,
                )
                for cat_feat_emb in self.categorical_feature_embeddings
            }
        )

    def _build_core_sequential_layers(self) -> None:
        self.core_sequential_layers = nn.Sequential(
            *[
                nn.Sequential(
                    BiLinearSemiNonNeg(
                        in_features_non_neg=in_monotone,
                        in_features_others=in_non_monotone,
                        out_features_non_neg=out_monotone,
                        out_features_others=out_non_monotone,
                        non_neg_inputs_name=self._monotone_feat_name,
                        other_inputs_name=self._non_monotone_feat_name,
                    ),
                    Partitioned(
                        **{
                            self._monotone_feat_name: self.activation_layer_monotone(),
                            self._non_monotone_feat_name: self.activation_layer_non_monotone(),
                        }
                    ),
                    Partitioned(
                        **{
                            self._monotone_feat_name: nn.Dropout(p=self.dropout_rate_monotone),
                            self._non_monotone_feat_name: nn.Dropout(
                                p=self.dropout_rate_non_monotone
                            ),
                        }
                    ),
                    Partitioned(
                        **{
                            self._monotone_feat_name: self.normalization_layer_monotone(
                                out_monotone
                            ),
                            self._non_monotone_feat_name: self.normalization_layer_non_monotone(
                                out_non_monotone
                            ),
                        }
                    ),
                )
                for in_non_monotone, out_non_monotone, in_monotone, out_monotone in zip(
                    [self.size_real_features_non_monotone + self.size_embeddings]
                    + self.hidden_sizes_non_monotone[:-1],
                    self.hidden_sizes_non_monotone,
                    [self.size_real_features_monotone] + self.hidden_sizes_monotone[:-1],
                    self.hidden_sizes_monotone,
                )
            ]
        )

    def _build_missing_layers(self) -> None:
        # Monotone layers
        nb_missing_layers_monotone = max(
            len(self.hidden_sizes_monotone) - len(self.core_sequential_layers), 0
        )
        self.missing_sequential_layers_monotone = nn.Sequential(
            *[
                nn.Sequential(
                    LinearNonNeg(in_features=in_features, out_features=out_features),
                    self.activation_layer_monotone(),
                    nn.Dropout(p=self.dropout_rate_monotone),
                    self.normalization_layer_monotone(out_features),
                )
                for in_features, out_features in zip(
                    self.hidden_sizes_monotone[-nb_missing_layers_monotone - 1 : -1],
                    self.hidden_sizes_monotone[-nb_missing_layers_monotone:],
                )
            ]
        )

        # Non_monotone layers
        nb_missing_layers_non_monotone = max(
            len(self.hidden_sizes_non_monotone) - len(self.core_sequential_layers), 0
        )
        self.missing_sequential_layers_non_monotone = nn.Sequential(
            *[
                nn.Sequential(
                    nn.Linear(in_features=in_features, out_features=out_features),
                    self.activation_layer_non_monotone(),
                    nn.Dropout(p=self.dropout_rate_non_monotone),
                    self.normalization_layer_non_monotone(out_features),
                )
                for in_features, out_features in zip(
                    self.hidden_sizes_non_monotone[-nb_missing_layers_non_monotone - 1 : -1],
                    self.hidden_sizes_non_monotone[-nb_missing_layers_non_monotone:],
                )
            ]
        )

    def _build_final_layer(self) -> None:
        self.final_linear_layer = LinearSemiNonNeg(
            in_features_non_neg=(
                self.hidden_sizes_monotone[-1]
                if len(self.hidden_sizes_monotone) > 0
                else self.size_real_features_monotone
            ),
            in_features_others=(
                self.hidden_sizes_non_monotone[-1]
                if len(self.hidden_sizes_non_monotone) > 0
                else self.size_real_features_non_monotone + self.size_embeddings
            ),
            out_features=1,
            non_neg_inputs_name=self._monotone_feat_name,
            other_inputs_name=self._non_monotone_feat_name,
        )
        self.final_sigmoid_layer = nn.Sigmoid()

    def probability_mapping(
        self,
        data_module: AugmentedBernoulliDataModule,
        other_features: pd.DataFrame,
        min_value: float,
        max_value: float,
        nb_points: int = 100,
        trainer: Optional[pl.Trainer] = None,
    ) -> Tuple[pd.DataFrame, np.ndarray, torch.Tensor]:
        """
        Computes the mapping between covariate `data_module.augmented_col` and the predicted probability
        of the neural network on the closed interval `[min_value, max_value]`,
        all `other_features` being fixed.
        The mapping is discretized into `nb_points` points.

        There are two ways to use this function:
        - call with `trainer=None` => method `self.predict` is used directly for
        inference
        - call with `trainer=pytorch_lightning.Trainer(...)` => method `predict` of
        `pytorch_lightning.Trainer` is used for inference
        """
        other_features = other_features.copy()  # make a copy before making modifications
        nb_samples = len(other_features)
        real_feature_values = np.array(
            [min_value + (max_value - min_value) * i / (nb_points - 1) for i in range(nb_points)],
            dtype=np.float32,
        )
        other_features[data_module.augmented_col] = np.array(1, dtype=np.float32)

        if trainer:
            data_module.prediction_df = other_features
            data_module.prediction_scaling_factors = real_feature_values
            self.prediction_min_augmented_value = min_value
            self.prediction_max_augmented_value = max_value
            bb_conversion_proba = torch.cat(trainer.predict(self, data_module))  # type: ignore
        else:
            bb_conversion_proba = self.predict_from_pandas(
                features=other_features,
                data_module=data_module,
                scaling_factors=real_feature_values,
                min_augmented_value=min_value,
                max_augmented_value=max_value,
            )
        other_features.drop(columns=[data_module.augmented_col], inplace=True)

        return (
            other_features,
            real_feature_values,
            torch.transpose(bb_conversion_proba.view(nb_points, nb_samples), dim0=0, dim1=1),
        )

    def assert_monotone_probability(
        self,
        data_module: AugmentedBernoulliDataModule,
        other_features: pd.DataFrame,
        non_decreasing: bool,
        min_value: float,
        max_value: float,
        nb_points: int = 100,
        trainer: Optional[pl.Trainer] = None,
        error_message: str = "",
    ):
        (
            other_features,
            real_feature_values,
            bb_conversion_proba,
        ) = self.probability_mapping(
            data_module=data_module,
            other_features=other_features,
            min_value=min_value,
            max_value=max_value,
            nb_points=nb_points,
            trainer=trainer,
        )

        assert_monotone(
            inputs=real_feature_values,
            outputs=bb_conversion_proba.numpy(force=True),
            non_decreasing=non_decreasing,
            error_message=error_message,
        )

    def plot_probability_mapping(
        self,
        data_module: AugmentedBernoulliDataModule,
        other_features: pd.DataFrame,
        min_value: float,
        max_value: float,
        nb_points: int = 100,
        x_title: str = "Covariate",
        y_title: str = "Predicted probability",
        title: str = "Evolution of the predicted probability as a function of the covariate",
        trainer: Optional[pl.Trainer] = None,
    ) -> plotly.graph_objects.Figure:
        """
        Plots the mapping between covariate `data_module.augmented_col` and the predicted probability
        of the neural network on the closed interval `[min_value, max_value]`,
        all `other_features` being fixed.
        The mapping is discretized into `nb_points` points.

        There are two ways to use this function:
        - call with `trainer=None` => method `self.predict` is used directly for
        inference
        - call with `trainer=pytorch_lightning.Trainer(...)` => method `predict` of
        `pytorch_lightning.Trainer` is used for inference
        """
        (
            other_features,
            real_feature_values,
            bb_conversion_proba,
        ) = self.probability_mapping(
            data_module=data_module,
            other_features=other_features,
            min_value=min_value,
            max_value=max_value,
            nb_points=nb_points,
            trainer=trainer,
        )

        nb_samples = len(other_features)
        informative_features = [
            f for f in data_module.output_features if f not in [data_module.augmented_col]
        ]
        other_features.index.names = ["Sample"]
        df = other_features.loc[
            np.repeat(other_features.index, nb_points), informative_features
        ].reset_index(drop=False)  # , names="Sample"
        df[x_title] = list(real_feature_values) * nb_samples
        df[y_title] = bb_conversion_proba.reshape(-1).numpy(force=True)

        fig = px.line(
            df,
            x=x_title,
            y=y_title,
            color="Sample",
            title=title,
            hover_data=informative_features,
        )
        fig.update_layout(hovermode="closest")
        return fig


Module = TypeVar("Module", bound=MonotoneBernoulliProbability)
DataModule = TypeVar("DataModule", bound=AugmentedBernoulliDataModule)


@dataclass(frozen=True)
class ProbabilityPredictor(Generic[Module, DataModule]):
    """
    Just a pair `(MonotoneBernoulliProbability, AugmentedBernoulliDataModule)`
    with useful methods such as `predict_from_pandas`.
    """

    module: Module
    data_module: DataModule

    @classmethod
    def load_from_checkpoint(
        cls,
        checkpoint_path: str,
        module_class: Type[Module] = MonotoneBernoulliProbability,
        data_module_class: Type[DataModule] = AugmentedBernoulliDataModule,  # type: ignore
        clear_data: bool = False,  # allows to save memory (clears training/validation/test data)
        compile_module: bool = False,  # allows to speed up inference on GPU
        compilation_kwargs: Dict[str, Any] = {},
        **kwargs,
    ) -> ProbabilityPredictor:
        data_module = data_module_class.load_from_checkpoint(checkpoint_path, **kwargs)
        module = module_class.load_from_checkpoint(checkpoint_path, **kwargs)
        proba_predictor = cls(
            module=torch.compile(module, **compilation_kwargs) if compile_module else module,
            data_module=data_module,
        )
        return proba_predictor.clear_data() if clear_data else proba_predictor

    def clear_data(self) -> ProbabilityPredictor[Module, DataModule]:
        self.data_module.clear_data()
        return self

    def predict_from_pandas(
        self,
        context: pd.DataFrame,
    ) -> np.ndarray:
        return self.module.predict(
            self.data_module.transform_to_tensors(context)  # type: ignore
        ).numpy(force=True)



from typing import Any, Literal, Optional, Tuple, Union

import torch
from sklearn.calibration import CalibrationDisplay
from torchmetrics import MeanAbsoluteError, MeanSquaredError
from torchmetrics.classification import BinaryCalibrationError
from torchmetrics.utilities.data import dim_zero_cat


class WeightedMeanSquaredError(MeanSquaredError):
    """
    Analogue of `torchmetrics.MeanSquaredError` but with (optional) sample weights.
    """

    def __init__(
        self,
        squared: bool = True,
        **kwargs: Any,
    ) -> None:
        super(MeanSquaredError, self).__init__(**kwargs)

        self.add_state("sum_squared_error", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.squared = squared

    def update(
        self,
        preds: torch.Tensor,
        target: torch.Tensor,
        sample_weights: Optional[torch.Tensor] = None,
    ) -> None:  # type: ignore
        """Update state with predictions and targets."""
        preds = preds if preds.is_floating_point is not None else preds.float()
        target = target if target.is_floating_point is not None else target.float()
        diff = preds - target
        sum_squared_error = torch.sum(
            diff * diff * (sample_weights if sample_weights is not None else 1)
        )
        self.sum_squared_error += sum_squared_error
        self.total += torch.sum(sample_weights) if sample_weights is not None else target.numel()


class WeightedMeanAbsoluteError(MeanAbsoluteError):
    """
    Analogue of `torchmetrics.MeanAbsoluteError` but with (optional) sample weights.
    """

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        self.add_state("sum_abs_error", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0.0), dist_reduce_fx="sum")

    def update(
        self,
        preds: torch.Tensor,
        target: torch.Tensor,
        sample_weights: Optional[torch.Tensor] = None,
    ) -> None:  # type: ignore
        """Update state with predictions and targets."""
        preds = preds if preds.is_floating_point is not None else preds.float()
        target = target if target.is_floating_point is not None else target.float()
        diff = preds - target
        sum_abs_error = torch.sum(
            torch.abs(diff) * (sample_weights if sample_weights is not None else 1)
        )
        self.sum_abs_error += sum_abs_error
        self.total += torch.sum(sample_weights) if sample_weights is not None else target.numel()


class WeightedBinaryCalibrationError(BinaryCalibrationError):
    """
    Analogue of `torchmetrics.classification import BinaryCalibrationError` but with (optional) sample weights.
    """

    def __init__(
        self,
        n_bins: int = 15,
        norm: Literal["l1", "l2", "max"] = "l1",
        ignore_index: Optional[int] = None,
        validate_args: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            n_bins=n_bins,
            norm=norm,
            ignore_index=ignore_index,
            validate_args=validate_args,
            **kwargs,
        )
        self.add_state("weights", [], dist_reduce_fx="cat")

    def update(
        self,
        preds: torch.Tensor,
        target: torch.Tensor,
        weights: Optional[torch.Tensor] = None,
    ) -> None:  # type: ignore
        super().update(preds=preds, target=target)
        if weights is not None:
            self.weights.append(weights)  # type: ignore

    def compute(self) -> torch.Tensor:
        confidences = dim_zero_cat(self.confidences)  # type: ignore
        accuracies = dim_zero_cat(self.accuracies)  # type: ignore
        weights = dim_zero_cat(self.weights) if self.weights else None  # type: ignore

        return self._ce_compute(
            confidences=confidences,
            accuracies=accuracies,
            weights=weights,
            bin_boundaries=self.n_bins,
            norm=self.norm,
        )

    def _ce_compute(
        self,
        confidences: torch.Tensor,
        accuracies: torch.Tensor,
        weights: Optional[torch.Tensor],
        bin_boundaries: Union[torch.Tensor, int],
        norm: str = "l1",
        debias: bool = False,
    ) -> torch.Tensor:
        """
        Analogue of `torchmetrics.functional.classification.calibration_error._ce_compute` but with weights.
        Computes the calibration error given the provided bin boundaries and norm.
        Args:
            confidences: The confidence (i.e. predicted prob) of the top1 prediction.
            accuracies: 1.0 if the top-1 prediction was correct, 0.0 otherwise.
            bin_boundaries: Bin boundaries separating the ``linspace`` from 0 to 1.
            norm: Norm function to use when computing calibration error. Defaults to "l1".
            debias: Apply debiasing to L2 norm computation as in
                `Verified Uncertainty Calibration`_. Defaults to False.
        Raises:
            ValueError: If an unsupported norm function is provided.
        Returns:
            Tensor: Calibration error scalar.
        """
        if isinstance(bin_boundaries, int):
            bin_boundaries = torch.linspace(
                0, 1, bin_boundaries + 1, dtype=torch.float, device=confidences.device
            )

        if norm not in {"l1", "l2", "max"}:
            raise ValueError(f"Norm {norm} is not supported. Please select from l1, l2, or max. ")  # nosec B608

        with torch.no_grad():
            acc_bin, conf_bin, prop_bin = self._binning_bucketize(
                confidences, accuracies, weights, bin_boundaries
            )

        if norm == "l1":
            ce = torch.sum(torch.abs(acc_bin - conf_bin) * prop_bin)
        elif norm == "max":
            ce = torch.max(torch.abs(acc_bin - conf_bin))
        elif norm == "l2":
            ce = torch.sum(torch.pow(acc_bin - conf_bin, 2) * prop_bin)
            # NOTE: debiasing is disabled in the wrapper functions. This implementation differs from that in sklearn.
            if debias:
                # the order here (acc_bin - 1 ) vs (1 - acc_bin) is flipped from
                # the equation in Verified Uncertainty Prediction (Kumar et al 2019)/
                debias_bins = (acc_bin * (acc_bin - 1) * prop_bin) / (
                    prop_bin * accuracies.size()[0] - 1
                )
                ce += torch.sum(
                    torch.nan_to_num(debias_bins)
                )  # replace nans with zeros if nothing appeared in a bin
            ce = torch.sqrt(ce) if ce > 0 else torch.tensor(0)
        return ce

    def _binning_bucketize(
        self,
        confidences: torch.Tensor,
        accuracies: torch.Tensor,
        weights: Optional[torch.Tensor],
        bin_boundaries: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Analogue of `torchmetrics.functional.classification.calibration_error._binning_bucketize` but with weights.
        Compute calibration bins using ``torch.bucketize``. Use for pytorch >= 1.6.
        Args:
            confidences: The confidence (i.e. predicted prob) of the top1 prediction.
            accuracies: 1.0 if the top-1 prediction was correct, 0.0 otherwise.
            bin_boundaries: Bin boundaries separating the ``linspace`` from 0 to 1.
        Returns:
            tuple with binned accuracy, binned confidence and binned probabilities
        """
        accuracies = accuracies.to(dtype=confidences.dtype)
        acc_bin = torch.zeros(
            len(bin_boundaries) - 1, device=confidences.device, dtype=confidences.dtype
        )
        conf_bin = torch.zeros(
            len(bin_boundaries) - 1, device=confidences.device, dtype=confidences.dtype
        )
        # count_bin = torch.zeros(len(bin_boundaries) - 1, device=confidences.device, dtype=confidences.dtype)
        weights_bin = torch.zeros(
            len(bin_boundaries) - 1,
            device=confidences.device,
            dtype=confidences.dtype if weights is None else weights.dtype,
        )

        indices = torch.clamp(
            torch.bucketize(confidences, bin_boundaries, right=True) - 1,
            min=0,
            max=len(weights_bin) - 1,
        )

        # count_bin.scatter_add_(dim=0, index=indices, src=torch.ones_like(confidences))
        weights_bin.scatter_add_(
            dim=0,
            index=indices,
            src=torch.ones_like(confidences) if weights is None else weights,
        )

        conf_bin.scatter_add_(
            dim=0, index=indices, src=confidences * (1 if weights is None else weights)
        )
        # conf_bin = torch.nan_to_num(conf_bin / count_bin)
        conf_bin = torch.nan_to_num(conf_bin / weights_bin)

        acc_bin.scatter_add_(
            dim=0, index=indices, src=accuracies * (1 if weights is None else weights)
        )
        # acc_bin = torch.nan_to_num(acc_bin / count_bin)
        acc_bin = torch.nan_to_num(acc_bin / weights_bin)

        # prop_bin = count_bin / count_bin.sum()
        prop_bin = weights_bin / weights_bin.sum()

        # Backup acc_bin, conf_bin, prop_bin to be used in calibration_curve method
        self.acc_bin_backup = acc_bin
        self.conf_bin_backup = conf_bin
        self.prop_bin = prop_bin

        return acc_bin, conf_bin, prop_bin

    def calibration_curve(self, estimator_name: str, pos_label: int = 1) -> CalibrationDisplay:
        return (
            CalibrationDisplay(
                prob_true=self.acc_bin_backup[self.prop_bin > 0].numpy(force=True),
                prob_pred=self.conf_bin_backup[self.prop_bin > 0].numpy(force=True),
                y_prob=None,
                estimator_name=estimator_name,
                pos_label=pos_label,
            )
            .plot()
            .figure_
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Literal, Optional

import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted
from torch.utils.data import Dataset

from pytorch_utils.pandas.utils import DataFrameRow, ListDataFrameRows
from pytorch_utils.dataset_configurations import (
    AugmentedBernoulliDatasetConfig,
    DataAugmentationConfig,
)


class MLStage(Enum):
    fit = auto()
    validate = auto()
    test = auto()
    predict = auto()


@dataclass(frozen=True)
class AugmentedBernoulliDataset(Dataset):
    """
    Implements storage-efficient data augmentation of Bernoulli samples
    (binary outcomes: successful or not) as well as data transformations
    (e.g., scaling, encoding, ...).
    If `is_success` is set to None, only features are generated (labels are dropped).
    This is useful for prediction sets.
    """

    data: ListDataFrameRows
    is_success: Optional[
        bool
    ]  # if None then only features are generated, if True labels are set to 1, If False labels are set to 0
    augmented_col: str
    fitted_preprocessing_pipeline: Optional[Pipeline] = None
    data_augmentation_scaling_factors: np.ndarray = np.array([1.0])
    label_col: str = "success_labels"  # name of label column
    labels_dtype: np.dtype = np.dtype("int32")
    sample_weight_col: Optional[str] = None
    min_augmented_value: float = -float("inf")
    max_augmented_value: float = float("inf")

    def __post_init__(self) -> None:
        # Sort data_augmentation scaling factors
        # This ensures that two dataclasses that only differ by the ordering of
        # `data_augmentation_scaling_factors` are considered equal
        object.__setattr__(
            self,
            "data_augmentation_scaling_factors",
            sorted(self.data_augmentation_scaling_factors),
        )

        # Check attributes
        if self.augmented_col not in self.data.columns:
            raise ValueError("`augmented_col` should be a column of dataframe `data`")

        if self.fitted_preprocessing_pipeline:
            check_is_fitted(
                self.fitted_preprocessing_pipeline,
                msg="fitted_preprocessing_pipeline is not fitted",
            )

        if len(self.data_augmentation_scaling_factors) == 0:
            raise ValueError("`data_augmentation_scaling_factors` should not be an empty array")

        # Compute augmented_df_indices and augmented_df_lengths_cumsum based on data_augmentation_scaling_factors.
        object.__setattr__(self, "augmented_df_indices", self._build_augmented_df_indices())
        object.__setattr__(
            self,
            "augmented_df_lengths_cumsum",
            self._build_augmented_df_lengths_cumsum(),
        )
        object.__setattr__(
            self,
            "indices_rows_mapping",
            {row.index: i for i, row in enumerate(self.data)},
        )

        # Set pipeline ouput to pandas
        if self.fitted_preprocessing_pipeline:
            self.fitted_preprocessing_pipeline.set_output(transform="pandas")

    @classmethod
    def from_config(
        cls,
        config: AugmentedBernoulliDatasetConfig,
        ml_stage: Literal[MLStage.fit, MLStage.validate, MLStage.test],
        fitted_preprocessing_pipeline: Optional[Pipeline] = None,
        label_col: str = "success_labels",
        labels_dtype: np.dtype = np.dtype("int32"),
        sample_weight_col: Optional[str] = None,
    ) -> AugmentedBernoulliDataset:
        if ml_stage not in [MLStage.fit, MLStage.validate, MLStage.test]:
            raise ValueError(
                "ml_stage should be one of MLStage.fit, MLStage.validate or MLStage.test"
            )

        return cls(
            data=ListDataFrameRows.from_pandas_df(
                config.training_data
                if ml_stage is MLStage.fit
                else config.validation_data
                if ml_stage is MLStage.validate
                else config.test_data
                if ml_stage is MLStage.test
                else None
            ),
            is_success=config.is_success,
            augmented_col=config.augmented_col,
            fitted_preprocessing_pipeline=fitted_preprocessing_pipeline,
            data_augmentation_scaling_factors=config.data_augmentation_scaling_factors,
            label_col=label_col,
            labels_dtype=labels_dtype,
            sample_weight_col=sample_weight_col,
            min_augmented_value=config.data_augmentation_config.min_value,
            max_augmented_value=config.data_augmentation_config.max_value,
        )

    @property
    def dataframe(self) -> pd.DataFrame:
        return self.data.dataframe

    def clear_data(self) -> AugmentedBernoulliDataset:
        object.__setattr__(self, "data", self.data[:0])
        return self

    def _build_augmented_df_indices(self) -> List[pd.Index]:
        df = self.data.build_dataframe([self.augmented_col])
        return [
            df[
                DataAugmentationConfig.scaling_filter(
                    df,
                    scaling_factor,
                    self.augmented_col,
                    min_value=self.min_augmented_value,
                    max_value=self.max_augmented_value,
                )
            ].index
            for scaling_factor in self.data_augmentation_scaling_factors
        ]

    def _build_augmented_df_lengths_cumsum(self) -> np.ndarray:
        return np.cumsum([len(indices) for indices in getattr(self, "augmented_df_indices")])

    @property
    def raw_feature_names(self):
        return self.data.columns

    @property
    def transformed_feature_names(self):
        return (
            self.fitted_preprocessing_pipeline.get_feature_names_out()
            if self.fitted_preprocessing_pipeline
            else self.data.columns
        )

    def __getitem__(self, index) -> DataFrameRow:
        """
        Implicit assumption in the following implementation: the preprocessing pipeline does not
        modify the number of rows.
        """

        if index > len(self) - 1:
            raise IndexError("single positional indexer is out-of-bounds")

        # Compute actual index
        # -------------------
        augmented_df_id = np.searchsorted(
            getattr(self, "augmented_df_lengths_cumsum"), index, side="right"
        )
        augmented_df_index = (
            index - getattr(self, "augmented_df_lengths_cumsum")[augmented_df_id - 1]
            if augmented_df_id > 0
            else index
        )
        actual_index = getattr(self, "augmented_df_indices")[augmented_df_id][augmented_df_index]
        row_number = getattr(self, "indices_rows_mapping")[actual_index]

        # Get raw features
        # -------------------
        raw_features: DataFrameRow = self.data[row_number]

        # Apply scaling factor
        # -------------------
        scaling_factor = self.data_augmentation_scaling_factors[augmented_df_id]
        raw_features = raw_features.set(
            column=self.augmented_col,
            value=scaling_factor * raw_features[self.augmented_col],
        )

        # Apply transformations
        # -------------------
        sample = (
            DataFrameRow.from_single_row_df(
                self.fitted_preprocessing_pipeline.transform(raw_features.single_row_df)
            )
            if self.fitted_preprocessing_pipeline
            else raw_features
        )

        # Add label
        # -------------------
        if self.is_success is not None:
            sample = sample.set(
                column=self.label_col, value=self.is_success, dtype=self.labels_dtype
            )

        # Add weights back if removed by preprocessing pipeline
        # -------------------
        if (self.sample_weight_col is not None) and (self.sample_weight_col not in sample):
            sample = sample.set(
                column=self.sample_weight_col,
                value=raw_features[self.sample_weight_col],
                dtype=raw_features.dtypes[self.sample_weight_col],
            )

        return sample

    def __len__(self):
        return self.augmented_df_lengths_cumsum[-1]



from __future__ import annotations

import copy
import itertools
from collections import UserList
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type

import numpy as np
import pandas as pd

# TYPE_CHECKING is False at runtime, but treated as True by type checkers.
# So the Python interpreters won't do pyspark import, but the type checker
# will still understand where it came from!
if TYPE_CHECKING:
    import pyspark.sql

    # We do not want to import pyspark at runtime, only when type checking (when building this project).
    # Because we want to get rid as much as possible of unnecessary (heavy) dependencies
    # in order to avoid having to install them too when installing this ML project (ex: for the inference part -> API).

from pytorch_utils.exceptions import InconsistentDatasetConfigurations
from pytorch_utils.logging.loggers import (
    Logger,
    VoidLogger,
    SingleLoggerDataclassLoggable,
    use_loggers,
)
from pytorch_utils.pandas.data_formatting import (
    PandasFormatter,
    PandasIdentityFormatter,
)
# from pytorch_utils.io.interface import MetaDataFrame


@dataclass(frozen=True)
class DataSplitConfig(SingleLoggerDataclassLoggable):
    """
    Configuration used to specify train, validation and test splits.
    The proportions should all lie in [0,1], with their sum smaller or equal to 1.
    In case the sum is strictly smaller than 1, only a random subset of the data is used.

    Attributes
    ----------
    training_proportion : float, default=1.
        proportion of dataframe to be used as training samples
    validation_proportion : float, default=0.
        proportion of dataframe to be used as validation samples
    test_proportion : float, default=0.
        proportion of dataframe to be used as test samples
    random_seed : Optional[int], default=None
        random seed used for random splitting
    stratify : Optional[List[str]], default=None
        list of column names used to stratify the data (see also sklearn.model_selection.train_test_split)
    """

    training_proportion: float = 1.0
    validation_proportion: float = 0.0
    test_proportion: float = 0.0
    random_seed: Optional[int] = None
    stratify: Optional[List[str]] = None
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def __post_init__(self):
        # Check attribute values
        if not 0.0 <= self.training_proportion <= 1.0:
            raise ValueError("training_proportion should belong to unit interval [0,1]")
        if not 0.0 <= self.validation_proportion <= 1.0:
            raise ValueError("validation_proportion should belong to unit interval [0,1]")
        if not 0.0 <= self.test_proportion <= 1.0:
            raise ValueError("test_proportion should belong to unit interval [0,1]")
        if (
            not self.training_proportion + self.validation_proportion + self.test_proportion
            <= 1.0 + 1e-6
        ):
            raise ValueError("The sum of all proportions should be at most 1")

        # Sort columns used to stratify splits
        # This ensures that two dataclasses that only differ by the ordering of `stratify` are considered equal
        object.__setattr__(
            self,
            "stratify",
            sorted(self.stratify) if self.stratify is not None else self.stratify,
        )

        # Make this class comply with Loggable Protocol
        object.__setattr__(self, "loggers", {"logger": self.logger})

    def train_valid_test_split(self, df: pd.DataFrame) -> Tuple[pd.Index, pd.Index, pd.Index]:
        def grouped_df(df: pd.DataFrame) -> pd.DataFrame:
            return (
                df.assign(
                    **{
                        # In case some values in self.stratify are null,
                        # we need to specify dropna=False to pandas groupby method
                        # but this causes some aggregation function (like "sample") to raise
                        # a KeyError. So the alternative method we found is to first
                        # replace null values by values with the same dtype that are not
                        # already present. N.B: this will work for both strings and numeric types.
                        f"{col}_bis_temp": df[col].fillna(df[col].max() * 2)
                        for col in self.stratify
                    }
                ).groupby(by=[f"{col}_bis_temp" for col in self.stratify])
                if self.stratify
                else df
            )

        training_indices = (
            grouped_df(df)
            .sample(frac=self.training_proportion, random_state=self.random_seed)
            .index
        )

        val_test_data = df[~df.index.isin(training_indices)]

        validation_indices = (
            grouped_df(val_test_data)
            .sample(
                frac=min(
                    self.validation_proportion / (1 - self.training_proportion)
                    if self.validation_proportion > 0
                    else 0.0,
                    1.0,
                ),
                random_state=self.random_seed,
            )
            .index
        )

        test_data = df[~df.index.isin(training_indices.union(validation_indices))]

        test_indices = (
            grouped_df(test_data)
            .sample(
                frac=min(
                    self.test_proportion
                    / (1 - self.training_proportion - self.validation_proportion)
                    if self.test_proportion > 0
                    else 0.0,
                    1.0,
                ),
                random_state=self.random_seed,
            )
            .index
        )

        return training_indices, validation_indices, test_indices


@dataclass(frozen=True)
class DataAugmentationConfig(SingleLoggerDataclassLoggable):
    """
    Configuration used to specify data augmentation on specific column (`augmented_col`).
    The idea of this data augmentation is to duplicate the data several times, with only `augmented_col` changed
    by multiplying the original values with a scaling factor.

    This can be useful when there is some monotone relationship between a covariate (`augmented_col`) and the
    success of an event (0=success, 1=event).

    Attributes
    ----------
    augmented_col : str
        Name of column to augment
    scaling_factors : np.ndarray
        Numpy array of floats corresponding to scaling factors used for data augmentation
    """

    augmented_col: str
    scaling_factors: np.ndarray = np.array([1.0])
    min_value: float = -float("inf")
    max_value: float = float("inf")
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def __post_init__(self):
        # Sort scaling factors
        # This ensures that two dataclasses that only differ by the ordering of `scaling_factors` are considered equal
        object.__setattr__(self, "scaling_factors", np.sort(self.scaling_factors))

        # Make this class comply with Loggable Protocol
        object.__setattr__(self, "loggers", {"logger": self.logger})

    def augment_data(self, data: pd.DataFrame) -> pd.DataFrame:
        return pd.concat(
            [
                self.scale_col(
                    data,
                    scaling_factor,
                    self.augmented_col,
                    min_value=self.min_value,
                    max_value=self.max_value,
                )
                for scaling_factor in self.scaling_factors
            ],
            axis=0,
            ignore_index=True,
            sort=True,
        )

    @staticmethod
    def scaling_filter(
        data: pd.DataFrame,
        scaling_factor: np.ndarray,
        col: str,
        min_value: float = -float("inf"),
        max_value: float = float("inf"),
    ) -> pd.DataFrame:
        """
        Filter used to drop scaled values that are outside the range [min_value, max_value].
        See method `scale_col`.
        """
        return (data[col] * scaling_factor >= min_value) & (data[col] * scaling_factor <= max_value)

    @staticmethod
    def scale_col(
        data: pd.DataFrame,
        scaling_factor: np.ndarray,
        col: str,
        min_value: float = -float("inf"),
        max_value: float = float("inf"),
    ) -> pd.DataFrame:
        """
        Return pandas dataframe identical to `data` with column `col` scaled by `scaling_factor`.
        The scaled values that are outside the range [min_value, max_value] are dropped.
        """
        scale_filter = DataAugmentationConfig.scaling_filter(
            data, scaling_factor, col, min_value, max_value
        )
        output_df = data[scale_filter].copy()
        output_df[col] *= scaling_factor
        return output_df

    @staticmethod
    def scaling_length(
        data: pd.DataFrame,
        scaling_factor: np.ndarray,
        col: str,
        min_value: float = -float("inf"),
        max_value: float = float("inf"),
    ) -> int:
        """
        Length of the pandas dataframe obtained when calling method `scale_col` with the exact same input.
        The implementation does not require explicitly building the dataframe.
        """
        scale_filter = DataAugmentationConfig.scaling_filter(
            data, scaling_factor, col, min_value, max_value
        )
        return scale_filter.sum()

    def augmentation_length(self, data: pd.DataFrame) -> int:
        """
        Length of the pandas dataframe obtained when calling method `augment_data` with the exact same input.
        The implementation does not require explicitly building the dataframe.
        """
        return sum(
            [
                self.scaling_length(
                    data,
                    scaling_factor,
                    self.augmented_col,
                    min_value=self.min_value,
                    max_value=self.max_value,
                )
                for scaling_factor in self.scaling_factors
            ]
        )

    @use_loggers("logger")
    def log(self, logger: Logger) -> None:
        params = asdict(self)
        del params["logger"]
        logger.log_params(params)


@dataclass(frozen=True)
class AugmentedBernoulliDatasetConfig(SingleLoggerDataclassLoggable):
    """
    Dataset configuration for augmented Bernoulli samples (binary outcomes: successful or not).
    There are 2 ways to construct an instance of `AugmentedBernoulliDatasetConfig`:
        1) either by calling the constructor and passing a pandas dataframe (with optional meatada) as input
        2) or by calling `class method `from_meta_dataframe` and passing a delta table as input
    Method 1 is preferred for testing/debugging/prototyping while method 2 is preferred for
    production and traceable experimentations (clean metadata, etc...).

    Attributes
    ----------
        data : pandas.DataFrame
            The Pandas dataframe containing the data.
        is_success : bool
            Whether the samples correspond to successful events or not
        split_config : DataSplitConfig
            The configuration for splitting between train, validation and test
        data_augmentation_config : DataAugmentationConfig
            The configuration for data augmentation
        metadata : Dict[str, Any], default={}
            Any information regarding the source data that we wish to track/save
    """

    data: pd.DataFrame = field(repr=False)
    is_success: bool
    data_augmentation_config: DataAugmentationConfig
    split_config: DataSplitConfig = DataSplitConfig()
    metadata: Dict[str, Any] = field(default_factory=dict)
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def __post_init__(self):
        # Compute training, validation and test indices
        (
            training_indices,
            validation_indices,
            test_indices,
        ) = self.split_config.train_valid_test_split(self.data)
        # N.B: sorting pandas data according to all columns (sorted in alphabetic order) for reproducibility
        object.__setattr__(
            self, "data", self.data.sort_values(by=self.columns).reset_index(drop=True)
        )

        object.__setattr__(self, "training_indices", training_indices)
        object.__setattr__(self, "validation_indices", validation_indices)
        object.__setattr__(self, "test_indices", test_indices)

        # Make this class comply with Loggable Protocol
        object.__setattr__(self, "loggers", {"logger": self.logger})

    @classmethod
    def from_meta_dataframe(
        cls,
        meta_df: Any[pyspark.sql.DataFrame],
        is_success: bool,
        data_augmentation_config: DataAugmentationConfig,
        split_config: DataSplitConfig,
        logger: Logger = VoidLogger(),
        spark_filter: Optional[str] = None,
        pandas_formatter: PandasFormatter = PandasIdentityFormatter(),
    ) -> AugmentedBernoulliDatasetConfig:
        """
        Use this method to construct an instance of `AugmentedBernoulliDatasetConfig` directly from a delta table
        with proper metadata.

        Attributes
        ----------
        delta_table : delta.tables.DeltaTable
            The delta table containing the data.
            To use a previous version of the data, call `restoreToVersion(version: int)`
            on `delta_table` before passing it to `AugmentedBernoulliDatasetConfig`.
        is_success : bool
            Whether the samples correspond to successful events or not
        split_config : DataSplitConfig
            The configuration for splitting between train, validation and test
        data_augmentation_config : DataAugmentationConfig
            The configuration for data augmentation
        pandas_formatter: PandasFormatter
            Any formatting on pandas data (cast dtypes, etc...).
        """
        metadata = copy.deepcopy(meta_df.metadata)
        metadata["common_spark_filter"] = spark_filter
        metadata["pandas_formatter"] = str(pandas_formatter)
        spark_filter = spark_filter if spark_filter is not None else "True"

        dataset_config = cls(
            data=pandas_formatter.format(meta_df.read().filter(spark_filter).toPandas()),
            is_success=is_success,
            data_augmentation_config=data_augmentation_config,
            split_config=split_config,
            metadata=metadata,
            logger=logger,
        )
        # Set _pandas_formatter attribute to be able to log info later
        object.__setattr__(dataset_config, "_pandas_formatter", pandas_formatter)

        return dataset_config

    def sample(self, n: int, replace: bool = False):
        return self.data.sample(n, replace=replace)

    @property
    def training_data(self):
        return self.data.iloc[self.training_indices]

    @property
    def training_data_length(self):
        return len(self.training_indices)

    @property
    def validation_data(self):
        return self.data.iloc[self.validation_indices]

    @property
    def validation_data_length(self):
        return len(self.validation_indices)

    @property
    def test_data(self):
        return self.data.iloc[self.test_indices]

    @property
    def test_data_length(self):
        return len(self.test_indices)

    @property
    def augmented_data(self):
        return self.data_augmentation_config.augment_data(self.data)

    @property
    def augmented_data_length(self):
        return self.data_augmentation_config.augmentation_length(self.data)

    @property
    def augmented_training_data(self):
        return self.data_augmentation_config.augment_data(self.training_data)

    @property
    def augmented_training_data_length(self):
        return self.data_augmentation_config.augmentation_length(self.training_data)

    @property
    def augmented_validation_data(self):
        return self.data_augmentation_config.augment_data(self.validation_data)

    @property
    def augmented_validation_data_length(self):
        return self.data_augmentation_config.augmentation_length(self.validation_data)

    @property
    def augmented_test_data(self):
        return self.data_augmentation_config.augment_data(self.test_data)

    @property
    def augmented_test_data_length(self):
        return self.data_augmentation_config.augmentation_length(self.test_data)

    @property
    def augmented_col(self):
        return self.data_augmentation_config.augmented_col

    @property
    def data_augmentation_scaling_factors(self):
        return self.data_augmentation_config.scaling_factors

    @property
    def columns(self):
        return sorted(self.data.columns)

    @property
    def dtypes(self):
        return self.data.dtypes

    def __len__(self) -> int:
        return len(self.data)

    def clear_data(self) -> AugmentedBernoulliDatasetConfig:
        object.__setattr__(self, "data", self.data[:0])
        return self

    @use_loggers("logger")
    def log(self, logger: Logger) -> None:
        self.split_config.log()
        self.data_augmentation_config.log()
        logger.log_param("metadata", self.metadata)
        logger.log_param("is_success", self.is_success)
        logger.log_param("columns", self.columns)
        logger.log_metrics(
            {
                "total_length": len(self),
                "total_augmented_length": self.augmented_data_length,
                "training_data_length": self.training_data_length,
                "augmented_training_data_length": self.augmented_training_data_length,
                "validation_data_length": self.validation_data_length,
                "augmented_validation_data_length": self.augmented_validation_data_length,
                "test_data_length": self.test_data_length,
                "augmented_test_data_length": self.augmented_test_data_length,
            }
        )
        logger.log_pandas_artifact(self.dtypes.rename("data_dtypes").to_frame(), "data_dtypes")
        logger.log_pandas_artifact(self.data.describe(), "data_statistics")

        if hasattr(self, "_pandas_formatter"):
            _pandas_formatter: PandasFormatter = getattr(self, "_pandas_formatter")
            _pandas_formatter.log()


class AugmentedBernoulliDatasetConfigs(UserList):
    def __init__(
        self,
        augmented_bernoulli_dataset_configs=List[AugmentedBernoulliDatasetConfig],
        label_col: str = "success_labels",
        labels_dtype: Type[np.intc] = np.int32,
        sample_weight_col: Optional[str] = None,
        logger: Logger = VoidLogger(),
    ):
        self.data: List[AugmentedBernoulliDatasetConfig] = augmented_bernoulli_dataset_configs
        self.label_col = label_col
        self.labels_dtype = labels_dtype
        self.sample_weight_col = sample_weight_col
        self.check_compatibility()
        self.loggers = {"logger": logger}  # make this class comply with Loggable Protocol

    def check_compatibility(self) -> None:
        if len(self.data) == 0:
            raise InconsistentDatasetConfigurations(
                "`augmented_bernoulli_dataset_configs` should have at least one element"
            )

        elif len(self.data) > 1:
            if self.sample_weight_col and not all(
                [self.sample_weight_col in conf.columns for conf in self.data]
            ):
                raise InconsistentDatasetConfigurations(
                    "All instances of `AugmentedBernoulliDatasetConfig` should have column `sample_weight_col`."
                )

            if not all(
                [self.data[0].augmented_col == conf.augmented_col for conf in self.data[1:]]
            ):
                raise InconsistentDatasetConfigurations(
                    "All instances of `AugmentedBernoulliDatasetConfig` should have "
                    + "the same attribute `augmented_col`."
                )

            if not all([set(self.data[0].columns) == set(conf.columns) for conf in self.data[1:]]):
                raise InconsistentDatasetConfigurations(
                    "All instances of `AugmentedBernoulliDatasetConfig` should have the same `columns`."
                )

            if not all([(self.data[0].dtypes == conf.dtypes).all() for conf in self.data[1:]]):
                raise InconsistentDatasetConfigurations(
                    "All instances of `AugmentedBernoulliDatasetConfig` should have the same `dtypes`."
                )

    @property
    def augmented_col(self):
        return self.data[0].augmented_col

    @property
    def columns(self):
        return self.data[0].columns

    @property
    def dtypes(self):
        return self.data[0].dtypes

    def _concat_data(self, list_df: List[pd.DataFrame], list_labels: List[bool]) -> pd.DataFrame:
        output_df = pd.concat(
            [df for df in list_df if len(df) > 0],
            axis=0,
            sort=True,
        )
        output_df[self.label_col] = list(
            itertools.chain(*[[label] * len(df) for df, label in zip(list_df, list_labels)])
        )
        return output_df.astype({self.label_col: self.labels_dtype})

    def sample(self, n: int, replace: bool = False):
        return self._concat_data(
            [conf_df.sample(n, replace=replace) for conf_df in self.data],
            [conf_df.is_success for conf_df in self.data],
        ).sample(n, replace=replace)

    @property
    def all_data(self):
        return self._concat_data(
            [conf_df.data for conf_df in self.data],
            [conf_df.is_success for conf_df in self.data],
        )

    @property
    def all_data_length(self):
        return sum([len(conf_df) for conf_df in self.data])

    @property
    def all_training_data(self):
        return self._concat_data(
            [conf_df.training_data for conf_df in self.data],
            [conf_df.is_success for conf_df in self.data],
        )

    @property
    def all_training_data_length(self):
        return sum([conf_df.training_data_length for conf_df in self.data])

    @property
    def all_validation_data(self):
        return self._concat_data(
            [conf_df.validation_data for conf_df in self.data],
            [conf_df.is_success for conf_df in self.data],
        )

    @property
    def all_validation_data_length(self):
        return sum([conf_df.validation_data_length for conf_df in self.data])

    @property
    def all_test_data(self):
        return self._concat_data(
            [conf_df.test_data for conf_df in self.data],
            [conf_df.is_success for conf_df in self.data],
        )

    @property
    def all_test_data_length(self):
        return sum([conf_df.test_data_length for conf_df in self.data])

    @property
    def all_augmented_data(self):
        return self._concat_data(
            [conf_df.augmented_data for conf_df in self.data],
            [conf_df.is_success for conf_df in self.data],
        )

    @property
    def all_augmented_data_length(self):
        return sum([conf_df.augmented_data_length for conf_df in self.data])

    @property
    def all_augmented_training_data(self):
        return self._concat_data(
            [conf_df.augmented_training_data for conf_df in self.data],
            [conf_df.is_success for conf_df in self.data],
        )

    @property
    def all_augmented_training_data_length(self):
        return sum([conf_df.augmented_training_data_length for conf_df in self.data])

    @property
    def all_augmented_validation_data(self):
        return self._concat_data(
            [conf_df.augmented_validation_data for conf_df in self.data],
            [conf_df.is_success for conf_df in self.data],
        )

    @property
    def all_augmented_validation_data_length(self):
        return sum([conf_df.augmented_validation_data_length for conf_df in self.data])

    @property
    def all_augmented_test_data(self):
        return self._concat_data(
            [conf_df.augmented_test_data for conf_df in self.data],
            [conf_df.is_success for conf_df in self.data],
        )

    @property
    def all_augmented_test_data_length(self):
        return sum([conf_df.augmented_test_data_length for conf_df in self.data])

    def clear_data(self) -> AugmentedBernoulliDatasetConfigs:
        self.data = [d.clear_data() for d in self.data]
        self.check_compatibility()
        return self

    @use_loggers("logger")
    def log(self, logger: Logger) -> None:
        for conf in self.data:
            conf.log()

        logger.log_param("label_col", self.label_col)
        logger.log_param("sample_weight_col", self.sample_weight_col)
        logger.log_param("augmented_col", self.augmented_col)
        logger.log_param("columns", self.columns)
        logger.log_metrics(
            {
                "all_data_length": self.all_data_length,
                "all_augmented_data_length": self.all_augmented_data_length,
                "all_training_data_length": self.all_training_data_length,
                "all_augmented_training_data_length": self.all_augmented_training_data_length,
                "all_validation_data_length": self.all_validation_data_length,
                "all_augmented_validation_data_length": self.all_augmented_validation_data_length,
                "all_test_data_length": self.all_test_data_length,
                "all_augmented_test_data_length": self.all_augmented_test_data_length,
            }
        )
        logger.log_pandas_artifact(self.dtypes.rename("data_dtypes").to_frame(), "data_dtypes")



import os
from typing import Optional

import mlflow
import pandas as pd


def get_or_create_mlflow_experiment(exp_name):
    exp = mlflow.get_experiment_by_name(exp_name)
    return mlflow.get_experiment(mlflow.create_experiment(exp_name)) if exp is None else exp


def get_active_mlflow_run_id():
    active_run = mlflow.active_run()
    return active_run.info.run_id if active_run else None


def get_active_mlflow_experiment_id():
    active_run = mlflow.active_run()
    return active_run.info.experiment_id if active_run else None


def mlflow_log_pandas_artifact(
    df: pd.DataFrame, df_name: str, artifact_path: Optional[str] = None
) -> None:
    tmp_path_csv = f"/tmp/{df_name}.csv"
    df.reset_index().to_csv(tmp_path_csv, index=False)
    mlflow.log_artifact(tmp_path_csv, artifact_path=artifact_path)
    os.remove(tmp_path_csv)



import logging
import os
from functools import wraps
from contextlib import contextmanager
from dataclasses import asdict, dataclass
from typing import (
    Any,
    Callable,
    List,
    Dict,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)
from typing_extensions import Self

import mlflow
import pandas as pd

from pytorch_utils.miscellaneous import (
    class_full_name,
    DataclassType,
)
from .utils import get_active_mlflow_experiment_id, get_active_mlflow_run_id


LOGGING_DISABLED = False  # Useful to disable logging globally


def disable_logging():
    global LOGGING_DISABLED
    LOGGING_DISABLED = True


def enable_logging():
    global LOGGING_DISABLED
    LOGGING_DISABLED = False


def is_logging_disabled():
    global LOGGING_DISABLED
    return LOGGING_DISABLED


@runtime_checkable
class Logger(Protocol):
    def log_param(self, key: str, value: Any) -> Any:
        return None

    def log_params(self, params: Dict[str, Any]) -> None:
        return None

    def log_metric(self, key: str, value: float, step: Optional[int] = None) -> None:
        return None

    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None) -> None:
        return None

    def log_artifact(self, local_path: str, artifact_path: Optional[str] = None) -> None:
        return None

    def log_pandas_artifact(self, df: Union[pd.DataFrame, pd.Series], df_name: str) -> None:
        return None

    def set_tag(self, key: str, value: Any) -> None:
        return None

    def set_tags(self, tags: Dict[str, Any]) -> None:
        return None

    def sklearn_autolog(self, **kwargs) -> None:
        return None

    def pytorch_autolog(self, **kwargs) -> None:
        return None

    def autolog(self, **kwargs) -> None:
        return None

    @contextmanager
    def context(self) -> Any:
        yield None


class VoidLogger(Logger):
    """
    Does not log anything
    """


@dataclass
class PythonLogger(Logger):
    """
    Use `logging.basicConfig(level=..., filename=..., force=True)` to set the root logger level, filename, etc...
    See: https://docs.python.org/3/library/logging.html#logging.basicConfig
    """

    logger: logging.Logger = logging.getLogger()
    logging_level: int = logging.INFO

    def log_param(self, key: str, value: Any) -> None:
        self.logger.log(self.logging_level, f"Parameter {key}={value}")

    def log_params(self, params: Dict[str, Any]) -> None:
        for k, v in params.items():
            self.log_param(k, v)

    def log_metric(self, key: str, value: float, step: Optional[int] = None) -> None:
        self.logger.log(self.logging_level, f"Metric {key}={value}")

    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None) -> None:
        for k, v in metrics.items():
            self.log_param(k, v)

    def set_tag(self, key: str, value: Any) -> None:
        self.logger = self.logger.getChild(value)

    def set_tags(self, tags: Dict[str, Any]) -> None:
        for k, v in tags.items():
            self.set_tag(k, v)


class MLFlowLogger(Logger):
    def __init__(self, **mlflow_run_params) -> None:
        self.mlflow_run_params = mlflow_run_params

    def log_param(self, key: str, value: Any) -> None:
        mlflow.log_param(key, value)

    def log_params(self, params: Dict[str, Any]) -> None:
        mlflow.log_params(params)

    def log_metric(self, key: str, value: float, step: Optional[int] = None) -> None:
        mlflow.log_metric(key, value, step)

    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None) -> None:
        mlflow.log_metrics(metrics, step)

    def log_artifact(self, local_path: str, artifact_path: Optional[str] = None) -> None:
        mlflow.log_artifact(local_path, artifact_path)

    def log_pandas_artifact(
        self, df: pd.DataFrame, df_name: str, artifact_path: Optional[str] = None
    ) -> None:
        tmp_path_csv = f"/tmp/{df_name}.csv"
        df.reset_index().to_csv(tmp_path_csv, index=False)
        self.log_artifact(tmp_path_csv, artifact_path=artifact_path)
        os.remove(tmp_path_csv)

    def set_tag(self, key: str, value: Any) -> None:
        mlflow.set_tag(key, value)

    def set_tags(self, tags: Dict[str, Any]) -> None:
        mlflow.set_tags(tags)

    def sklearn_autolog(self, **kwargs) -> None:
        mlflow.sklearn.autolog(**kwargs)

    def pytorch_autolog(self, **kwargs) -> None:
        mlflow.pytorch.autolog(**kwargs)

    def autolog(self, **kwargs) -> None:
        mlflow.autolog(**kwargs)

    @contextmanager
    def context(self):
        with mlflow.start_run(**self.mlflow_run_params) as mlflow_run:
            yield mlflow_run


@dataclass
class ActiveRunMLFlowLogger(MLFlowLogger):
    nested: bool = False
    tags: Optional[Dict[str, Any]] = None
    description: Optional[str] = None

    @property
    def mlflow_run_params(self):
        run_params = asdict(self)
        run_params.update({"run_id": get_active_mlflow_run_id()})
        return run_params


@dataclass
class ActiveExperimentMLFlowLogger(MLFlowLogger):
    run_name: Optional[str] = None
    nested: bool = False
    tags: Optional[Dict[str, Any]] = None
    description: Optional[str] = None

    @property
    def mlflow_run_params(self):
        run_params = asdict(self)
        run_params.update({"experiment_id": get_active_mlflow_experiment_id()})
        return run_params


class Loggable(Protocol):
    loggers: Dict[str, Logger]


def use_loggers(*logger_names: str):
    def decorator(log_method: Callable):
        @wraps(log_method)
        def wrapper(self: Loggable, *args, **kwargs):
            if is_logging_disabled():
                log_method(self, logger=VoidLogger(), *args, **kwargs)
            else:
                for loger_name in logger_names:
                    logger = self.loggers[loger_name]
                    with logger.context():
                        log_method(self, logger=logger, *args, **kwargs)

        return wrapper

    return decorator


@runtime_checkable
class SingleLoggerDataclassLoggable(DataclassType, Protocol):
    """A specific ``Loggable``"""

    logger: Logger

    def __post_init__(self):
        self.set_logger(self.logger)

    def set_logger(self, logger: Logger) -> Self:
        object.__setattr__(self, "logger", logger)
        object.__setattr__(self, "loggers", {"logger": logger})
        return self

    @use_loggers("logger")
    def log(self, logger: Logger, params_not_to_log: List[str] = ["logger"]) -> None:
        params = asdict(self)
        params = {
            param_name: getattr(self, param_name) for param_name in params
        }  # by default, asdict explodes nested dataclasses as dict too

        for param in params.values():
            if isinstance(param, SingleLoggerDataclassLoggable):
                param.log(params_not_to_log=params_not_to_log)

        for param_name in params_not_to_log:
            del params[param_name]

        for param_name, param_value in params.items():
            if isinstance(param_value, pd.DataFrame):
                logger.log_pandas_artifact(
                    param_value,
                    df_name=param_name,
                )
                del params[param_name]

        logger.log_params(params)
        logger.log_param("type", class_full_name(self.__class__))
        logger.log_param("repr", repr(self))



from __future__ import annotations

import copy
from typing import Any, Dict, List, Literal, Optional, Set, Union, cast, no_type_check

import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
from pytorch_lightning.core.saving import _load_from_checkpoint
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted
from torch.utils.data import ConcatDataset, DataLoader, Dataset

from pytorch_utils.logging.loggers import (
    Logger,
    VoidLogger,
    is_logging_disabled,
    use_loggers,
)
from pytorch_utils.pandas.utils import DataFrameRow, ListDataFrameRows
from pytorch_utils.dataset_configurations import (
    AugmentedBernoulliDatasetConfigs,
    DataAugmentationConfig,
)
from pytorch_utils.datasets import (
    AugmentedBernoulliDataset,
    MLStage,
)
from pytorch_utils.utils import (
    BatchTorchTensors,
    NamedTorchTensors,
)


class AugmentedBernoulliDataModule(pl.LightningDataModule):
    """
    For prediction: attributes `prediction_df` (and
    optionally `prediction_scaling_factors`,
    `prediction_min_augmented_value` and
    `prediction_max_augmented_value`) must be set as desired.
    """

    def __init__(
        self,
        augmented_bernoulli_dataset_configs: AugmentedBernoulliDatasetConfigs,
        preprocessing_pipeline: Optional[Pipeline] = None,
        train_dataloader_params: Dict[str, Any] = {},
        val_dataloader_params: Dict[str, Any] = {},
        test_dataloader_params: Dict[str, Any] = {},
        predict_dataloader_params: Dict[str, Any] = {},
        data_module_logger: Logger = VoidLogger(),
        preprocessing_pipeline_logger: Logger = VoidLogger(),
        prepare_data_per_node: bool = False,
    ):
        super().__init__()

        # Set attributes
        # -------------------
        self.prepare_data_per_node = prepare_data_per_node

        self.augmented_bernoulli_dataset_configs = augmented_bernoulli_dataset_configs

        self.augmented_col = self.augmented_bernoulli_dataset_configs.augmented_col
        self.label_col = self.augmented_bernoulli_dataset_configs.label_col
        self.sample_weight_col = self.augmented_bernoulli_dataset_configs.sample_weight_col

        self.preprocessing_pipeline = preprocessing_pipeline

        # Copying the dataloader params before updating them is safer (it avoids
        # modifying the same underlying object when the same dictionary is passed
        # as input several times: see below with collate_fn and shuffle)
        self.train_dataloader_params = copy.deepcopy(train_dataloader_params)
        self.val_dataloader_params = copy.deepcopy(val_dataloader_params)
        self.test_dataloader_params = copy.deepcopy(test_dataloader_params)
        self.predict_dataloader_params = copy.deepcopy(predict_dataloader_params)

        self._batch_size: Optional[int] = None
        # see property batch_size and associated setter below

        self.loggers = {
            "data_module_logger": data_module_logger,
            "preprocessing_pipeline_logger": preprocessing_pipeline_logger,
        }

        # Teh following attributes must be set depending on what we wish to predict
        self.prediction_df: Optional[pd.DataFrame] = None
        self.prediction_scaling_factors: np.ndarray = np.array([1.0])
        self.prediction_min_augmented_value: float = -float("inf")
        self.prediction_max_augmented_value: float = float("inf")

        self._input_features_dtypes: Optional[pd.Series] = (
            None  # features dtypes before preprocessing
        )
        self._output_features_dtypes: Optional[pd.Series] = (
            None  # features dtypes after preprocessing
        )
        self._output_real_features: Optional[List[str]] = None
        self._output_categorical_features: Optional[Dict[str, Any]] = None

        self.prediction_set: Optional[Dataset] = None
        self.test_set: Optional[Dataset] = None
        self.validation_set: Optional[Dataset] = None
        self.train_set: Optional[Dataset] = None

        # Set pipeline outputs to pandas dataframes
        # -------------------
        if self.preprocessing_pipeline:
            self.preprocessing_pipeline.set_output(transform="pandas")

        # Overwrite collate_fn with pandas-specific version
        # -------------------
        self.train_dataloader_params.update({"collate_fn": self.preprocessing_pandas_collate_fn})
        self.val_dataloader_params.update({"collate_fn": self.preprocessing_pandas_collate_fn})
        self.test_dataloader_params.update({"collate_fn": self.preprocessing_pandas_collate_fn})
        self.predict_dataloader_params.update(
            {
                "shuffle": False,  # never shuffle when predicting
                "collate_fn": self.preprocessing_pandas_collate_fn,
            }
        )

        # Save hyperparameters
        # -------------------
        self.save_hyperparameters()  # saves all constructor params by default

    def preprocessing_pandas_collate_fn(
        self, batch: List[DataFrameRow]
    ) -> Union[NamedTorchTensors, BatchTorchTensors]:
        """
        More computationally efficient than using `collate_fn=lambda batch: pd.concat(batch, axis=0, sort=True)` and
        setting `fitted_transformers_pipeline=self.preprocessing_pipeline` in AugmentedBernoulliDataset. This
        vectorized the transform operations.
        """
        # collated_batch = pd.concat(batch, axis=0, sort=True)
        collated_batch = ListDataFrameRows.from_list_rows(batch).dataframe
        return self.transform_to_tensors(collated_batch)

    def transform(self, batch: pd.DataFrame) -> pd.DataFrame:
        """
        Implicit assumption in the following implementation: the preprocessing pipeline does not
        modify the number of rows.
        """
        contains_labels = self.label_col in batch
        contains_weights = (self.sample_weight_col is not None) and (
            self.sample_weight_col in batch
        )
        non_feature_cols = (
            []
            + ([self.label_col] if contains_labels else [])
            + ([self.sample_weight_col] if contains_weights else [])
        )
        features = batch.drop(columns=non_feature_cols)
        transformed_batch = (
            self.preprocessing_pipeline.transform(features)
            if self.preprocessing_pipeline
            else features
        )
        if contains_labels:
            transformed_batch[self.label_col] = batch[self.label_col]
        if contains_weights and (self.sample_weight_col not in transformed_batch):
            # add back sample_weight_col if removed by preprocessing pipeline
            transformed_batch[self.sample_weight_col] = batch[self.sample_weight_col]
        return transformed_batch

    def transform_to_tensors(
        self, batch: pd.DataFrame
    ) -> Union[NamedTorchTensors, BatchTorchTensors]:
        return self.format_to_tensors(self.transform(batch))

    def augment_transform_to_tensors(
        self,
        batch: pd.DataFrame,
        augmentation_scaling_factors: np.ndarray = np.array([1.0]),
        min_augmented_value: float = -float("inf"),
        max_augmented_value: float = float("inf"),
    ) -> Union[NamedTorchTensors, BatchTorchTensors]:
        bb_conv_data_augmentation_conf = DataAugmentationConfig(
            augmented_col=self.augmented_col,
            scaling_factors=augmentation_scaling_factors,
            min_value=min_augmented_value,
            max_value=max_augmented_value,
        )
        return self.transform_to_tensors(bb_conv_data_augmentation_conf.augment_data(batch))

    def format_to_tensors(
        self, transformed_batch: pd.DataFrame
    ) -> Union[NamedTorchTensors, BatchTorchTensors]:
        contains_labels = self.label_col in transformed_batch
        contains_weights = (self.sample_weight_col is not None) and (
            self.sample_weight_col in transformed_batch
        )
        non_feature_cols = (
            []
            + ([self.label_col] if contains_labels else [])
            + ([self.sample_weight_col] if contains_weights else [])
        )
        formatted_features = self._format_pandas_dataframe(
            transformed_batch.drop(columns=non_feature_cols)
        )
        if contains_labels:
            formatted_labels = self._format_pandas_series(transformed_batch[self.label_col])
            formatted_weights = (
                self._format_pandas_series(transformed_batch[self.sample_weight_col])
                if contains_weights
                else None
            )
            return (formatted_features, formatted_labels, formatted_weights)
        else:
            return formatted_features

    def _format_pandas_dataframe(self, dataframe: pd.DataFrame) -> NamedTorchTensors:
        return {
            f: torch.from_numpy(
                dataframe[
                    # 1D tensors for categorical features, 2D for all other features
                    f if f in self.output_categorical_features else [f]
                ].to_numpy()
            )
            for f in dataframe
        }

    def _format_pandas_series(self, series: pd.Series) -> torch.Tensor:
        return torch.from_numpy(series.to_numpy())

    @property
    def is_preprocessing_pipeline_fitted(self) -> bool:
        """
        Boolean indicating whether pipeline is fitted
        """
        if self.preprocessing_pipeline:
            try:
                check_is_fitted(self.preprocessing_pipeline)
                is_preprocessing_pipeline_fitted = True
            except NotFittedError:
                is_preprocessing_pipeline_fitted = False
            return is_preprocessing_pipeline_fitted
        else:
            return True

    def fit_preprocessing_pipeline(self, refit: bool = False) -> None:
        if (refit or not self.is_preprocessing_pipeline_fitted) and self.preprocessing_pipeline:
            self._fit_preprocessing_pipeline()

    @use_loggers("preprocessing_pipeline_logger")
    def _fit_preprocessing_pipeline(self, logger: Logger) -> None:
        """
        Fit training pipeline if not None (otherwise do nothing).
        """
        logger.sklearn_autolog(disable=is_logging_disabled())
        if self.preprocessing_pipeline is not None:
            self.preprocessing_pipeline.fit(
                self.augmented_bernoulli_dataset_configs.all_training_data.drop(
                    columns=[self.label_col]
                    + ([self.sample_weight_col] if self.sample_weight_col is not None else [])
                )
            )
        logger.sklearn_autolog(disable=True)

    @property
    def input_features_dtypes(self) -> pd.Series:
        if self._input_features_dtypes is None:
            self._input_features_dtypes = (
                self.augmented_bernoulli_dataset_configs.sample(10, replace=True)
                .drop(
                    columns=[self.label_col]
                    + ([self.sample_weight_col] if self.sample_weight_col is not None else [])
                )
                .dtypes
            )

        return self._input_features_dtypes

    @property
    def output_features_dtypes(self) -> pd.Series:
        if self._output_features_dtypes is None:
            self.fit_preprocessing_pipeline(refit=False)

            self._output_features_dtypes = self.transform(
                self.augmented_bernoulli_dataset_configs.sample(10, replace=True).drop(
                    columns=[self.label_col]
                    + ([self.sample_weight_col] if self.sample_weight_col is not None else [])
                )
            ).dtypes

        return self._output_features_dtypes

    @property
    def output_features(self):
        return sorted(self.output_features_dtypes.index.to_list())

    @property
    def output_real_features(self) -> List[str]:
        if self._output_real_features is None:
            self._output_real_features = sorted(
                self.output_features_dtypes[
                    self.output_features_dtypes.map(lambda x: np.issubdtype(x, np.floating))
                ].index
            )

        return self._output_real_features

    @property
    def output_categorical_features(self) -> Dict[str, Set[int]]:
        if self._output_categorical_features is None:
            categorical_features_names = sorted(
                self.output_features_dtypes[
                    self.output_features_dtypes.map(lambda x: np.issubdtype(x, np.integer))
                ].index
            )

            training_data = self.augmented_bernoulli_dataset_configs.all_training_data[
                categorical_features_names
            ]

            self._output_categorical_features = {
                f: training_data[f].unique() for f in categorical_features_names
            }

        return self._output_categorical_features

    def prepare_data(self) -> None:
        # Fit preprocessing pipeline
        # -------------------
        # Preferable to fit the pipeline in `prepare_data` rather than in `setup` since in case of
        # parallel data loading, the `setup` method will be called once in every thread which is useless
        self.fit_preprocessing_pipeline(refit=True)
        self.log()

    def setup(self, stage: str) -> None:
        try:
            MLStage[stage]
        except KeyError:
            raise NotImplementedError(
                f"Argument stage should be either 'fit', 'validate', 'test' or 'predict' but '{stage}' was given."
            )

        if stage == MLStage.predict.name and self.prediction_df is None:
            raise ValueError("Attribute prediction_df must be set before predict setup.")

        # Fit preprocessing pipeline
        self.fit_preprocessing_pipeline(refit=False)

        # Create dataset corresponding to ml stage
        dataset = (
            AugmentedBernoulliDataset(
                data=ListDataFrameRows.from_pandas_df(self.prediction_df),
                is_success=None,  # labels are unknown in this case
                augmented_col=self.augmented_col,
                fitted_preprocessing_pipeline=None,
                data_augmentation_scaling_factors=self.prediction_scaling_factors,
                label_col=self.label_col,
                labels_dtype=np.dtype("int32"),
                sample_weight_col=None,
                min_augmented_value=self.prediction_min_augmented_value,
                max_augmented_value=self.prediction_max_augmented_value,
            )
            if stage == MLStage.predict.name
            else (
                ConcatDataset(
                    [
                        AugmentedBernoulliDataset.from_config(
                            config=ds_config,
                            ml_stage=MLStage[stage],  # type: ignore
                            fitted_preprocessing_pipeline=None,
                            label_col=self.label_col,
                            labels_dtype=np.dtype("int32"),
                            sample_weight_col=None,
                        )
                        for ds_config in self.augmented_bernoulli_dataset_configs
                    ]
                )
            )
        )

        # For the fit stage, also create validation set
        validation_dataset: Optional[Dataset] = (
            ConcatDataset(
                [
                    AugmentedBernoulliDataset.from_config(
                        config=ds_config,
                        ml_stage=MLStage.validate,
                        fitted_preprocessing_pipeline=None,
                        label_col=self.label_col,
                        labels_dtype=np.dtype("int32"),
                        sample_weight_col=None,
                    )
                    for ds_config in self.augmented_bernoulli_dataset_configs
                ]
            )
            if stage == MLStage.fit.name
            else None
        )

        # Set dataset attribute
        self.setup_datasets(MLStage[stage], dataset, validation_dataset)

    def setup_datasets(
        self,
        ml_stage: Literal[MLStage.fit, MLStage.validate, MLStage.test, MLStage.predict],
        dataset: Dataset,
        validation_dataset: Optional[Dataset] = None,
    ) -> None:
        if ml_stage is MLStage.fit:
            self.training_set = dataset
            if validation_dataset is None:
                raise ValueError("validation_dataset should not be None when ml_stage=MLStage.fit")
            self.validation_set = validation_dataset
        elif ml_stage is MLStage.validate:
            self.validation_set = dataset
        elif ml_stage is MLStage.test:
            self.test_set = dataset
        elif ml_stage is MLStage.predict:
            self.prediction_set = dataset
        else:
            raise NotImplementedError(f"Stage {ml_stage} is not implemented")

    @property
    def batch_size(self):
        return self._batch_size

    @batch_size.setter
    def batch_size(self, batch_size: int) -> None:
        """
        Updates batch size (useful to apply `pytorch_lightning.trainer.Trainer.tune.scale_batch_size`)
        """
        self._batch_size = batch_size

        # update batch size of all dataloaders
        self.train_dataloader_params.update({"batch_size": batch_size})
        self.val_dataloader_params.update({"batch_size": batch_size})
        self.test_dataloader_params.update({"batch_size": batch_size})
        self.predict_dataloader_params.update({"batch_size": batch_size})

    def train_dataloader(self):
        return DataLoader(self.training_set, **self.train_dataloader_params)

    def val_dataloader(self):
        return DataLoader(self.validation_set, **self.val_dataloader_params)

    def test_dataloader(self):
        return DataLoader(self.test_set, **self.test_dataloader_params)

    def predict_dataloader(self):
        return DataLoader(self.prediction_set, **self.predict_dataloader_params)

    def clear_data(self):
        self.input_features_dtypes  # just to make sure features have been saved before erasing data
        self.output_categorical_features  # just to make sure categories have been computed before erasing data
        self.augmented_bernoulli_dataset_configs.clear_data()
        if hasattr(self, "training_set"):
            del self.training_set
        if hasattr(self, "validation_set"):
            del self.validation_set
        if hasattr(self, "test_set"):
            del self.test_set
        if hasattr(self, "prediction_set"):
            del self.prediction_set
        return self

    @use_loggers("data_module_logger")
    def log(self, logger: Logger) -> None:
        self.augmented_bernoulli_dataset_configs.log()

        logger.log_param("augmented_col", self.augmented_col)
        logger.log_param("label_col", self.label_col)
        logger.log_param("sample_weight_col", self.sample_weight_col)
        logger.log_param("prepare_data_per_node", self.prepare_data_per_node)
        logger.log_param("batch_size", self.batch_size)
        logger.log_params(
            {f"train_dataloader__{k}": v for k, v in self.train_dataloader_params.items()}
        )
        logger.log_params(
            {f"val_dataloader__{k}": v for k, v in self.val_dataloader_params.items()}
        )
        logger.log_params(
            {f"test_dataloader__{k}": v for k, v in self.test_dataloader_params.items()}
        )
        logger.log_params(
            {f"predict_dataloader__{k}": v for k, v in self.predict_dataloader_params.items()}
        )

        if self.is_preprocessing_pipeline_fitted:
            logger.log_param("real_features", self.output_real_features)
            logger.log_pandas_artifact(
                self.input_features_dtypes.to_frame(), "input_features_dtypes"
            )
            logger.log_pandas_artifact(
                self.output_features_dtypes.to_frame(), "output_features_dtypes"
            )
            categorical_features_cardinals = pd.Series(
                {k: len(v) for k, v in self.output_categorical_features.items()},
                name="cardinals",
            )
            logger.log_pandas_artifact(
                categorical_features_cardinals.to_frame(),
                "categorical_features_cardinals",
            )

    @no_type_check
    @classmethod
    def load_from_checkpoint(
        cls,
        checkpoint_path,
        map_location=None,
        hparams_file=None,
        strict=None,
        **kwargs: Any,
    ) -> AugmentedBernoulliDataModule:
        """
        We override this method to correct a bug with `map_location` argument.
        See Github issue: https://github.com/Lightning-AI/lightning/issues/17945
        """
        loaded = _load_from_checkpoint(
            cls,
            checkpoint_path,
            map_location=map_location,
            hparams_file=hparams_file,
            strict=strict,
            **kwargs,
        )
        return cast(AugmentedBernoulliDataModule, loaded)



from typing import Hashable, List, TypeVar, Protocol, ClassVar, Any, Dict


def class_full_name(cls):
    return ".".join([cls.__module__, cls.__name__])


T = TypeVar("T", bound=Hashable)


def sort_hashable(array: List[T]) -> List[T]:
    return sorted(array, key=lambda x: hash(x))


class DataclassType(Protocol):
    __dataclass_fields__: ClassVar[Dict[str, Any]]



from dataclasses import asdict, dataclass, field
from typing import Mapping, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
import torchmetrics

from pytorch_utils.exceptions import (
    NotMonotone,
    NotNonDecreasing,
    NotNonIncreasing,
)
from pytorch_utils.logging.loggers import (
    Logger,
    VoidLogger,
    use_loggers,
)

NamedTorchTensors = Mapping[str, torch.Tensor]
NamedTorchModules = Mapping[str, nn.Module]
NamedTorchMetrics = Mapping[str, torchmetrics.Metric]
BatchTorchTensors = Tuple[NamedTorchTensors, torch.Tensor, Optional[torch.Tensor]]


@dataclass(frozen=True, order=True)
class CategoricalFeatureEmbedding:
    """
    Collection that simplifies the constructor parameters.
    """

    feature_name: str
    nb_distinct_values: int
    embedding_size: int
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def __post_init__(self):
        # Check attributes
        # ------------------------
        if self.nb_distinct_values < 0:
            raise ValueError("nb_distinct_values must be non-negative")
        if self.embedding_size < 0:
            raise ValueError("embedding_size must be non-negative")

        # Make this class comply with Loggable Protocol
        object.__setattr__(self, "loggers", {"logger": self.logger})

    @use_loggers("logger")
    def log(self, logger: Logger) -> None:
        params = asdict(self)
        del params["logger"]
        logger.log_params(params)


def get_embedding_size(
    nb_categories: int,
    multiplicative_factor: float = 1.6,
    power_exponent: float = 0.56,
    max_size: int = 600,
) -> int:
    """
    Determine empirically good embedding sizes (formula taken from fastai: https://docs.fast.ai/tabular.model.html).
    Parameters
    ----------
    nb_categories: int
        number of categories
    max_size: (int, optional)
        maximum embedding size. Defaults to 600.
    Returns
    -------
    int
        embedding size
    """
    if nb_categories > 2:
        return min(round(multiplicative_factor * nb_categories**power_exponent), max_size)
    else:
        return 1


def assert_monotone(
    inputs: np.ndarray,
    outputs: np.ndarray,
    non_decreasing: Optional[bool] = None,
    error_message: str = "",
    tol: float = 1e-5,
) -> None:
    """
    Asserts if the `outputs` are a monontone function of the `inputs`.
    `inputs` should be a 1-dimensional.
    `outputs` should be a 1 or 2-dimensional array. If 2-dimensional,
    then every row is tested to be a monotone function of the inputs.
    If `non_decreasing` is `None` then an exception is raised only if the mapping
    is neither non-decreasing, nor non-increasing.
    If `non_decreasing` is `True` then an exception is raised only if the mapping
    is not non-decreasing.
    If `non_decreasing` is `False` then an exception is raised only if the mapping
    is not non-increasing.
    """
    if len(inputs.shape) != 1:
        raise ValueError("inputs should be a 1-dimensional array")

    if len(outputs.shape) not in [1, 2]:
        raise ValueError("outputs should be a 1 or 2-dimensional array")

    if inputs.shape[0] != outputs.shape[-1]:
        raise ValueError("the last dimension of outputs should match the dimension of inputs")

    argsort_indices = np.argsort(inputs)

    is_non_decreasing = np.all(np.diff(outputs[..., argsort_indices], n=1, axis=1) >= -tol)
    is_non_increasing = np.all(np.diff(outputs[..., argsort_indices], n=1, axis=1) <= tol)

    if (non_decreasing is None) and not (is_non_decreasing or is_non_increasing):
        raise NotMonotone(error_message)
    elif (non_decreasing is True) and not is_non_decreasing:
        raise NotNonIncreasing(error_message)
    elif (non_decreasing is False) and not is_non_increasing:
        raise NotNonDecreasing(error_message)



from __future__ import annotations

import functools
from dataclasses import dataclass, field
from typing import Any, List, Protocol

import pandas as pd

from pytorch_utils.logging.loggers import (
    Logger,
    VoidLogger,
    SingleLoggerDataclassLoggable,
    use_loggers,
)


def cast_float_cols(df: pd.DataFrame, target_dtype: str) -> pd.DataFrame:
    """
    Cast all float columns to a target dtype.
    """
    # Identify columns with float dtype
    casted_df = df.copy()
    float_cols = [col for col in casted_df.columns if pd.api.types.is_float_dtype(casted_df[col])]

    # Cast to float32
    casted_df[float_cols] = casted_df[float_cols].astype(target_dtype)
    return casted_df


def cast_integer_cols(df: pd.DataFrame, target_dtype: str) -> pd.DataFrame:
    """
    Cast all integer columns to a target dtype.
    """
    # Identify columns with float dtype
    casted_df = df.copy()
    integer_cols = [
        col for col in casted_df.columns if pd.api.types.is_integer_dtype(casted_df[col])
    ]

    # Cast to float32
    casted_df[integer_cols] = casted_df[integer_cols].astype(target_dtype)
    return casted_df


class PandasFormatter(SingleLoggerDataclassLoggable, Protocol):
    def format(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    def __mul__(self, other_formatter: PandasFormatter) -> PandasFormatter:
        """
        Composition function (https://en.wikipedia.org/wiki/Function_composition)
        Allows to easily build "composite" formaters.
        `formater2 * formater1` returns a `PandasFormater` with a format method that
        first applies the `format` method of `formater1` and then applies the `format`
        method of `formater2` to the resulting pandas dataframe.
        Note that this operation is not commutative (like matrix multiplication).
        """
        return CompositeFormatter(formatters=[other_formatter, self], logger=self.logger).flatten()


@dataclass
class CompositeFormatter(PandasFormatter):
    formatters: List[PandasFormatter]
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def format(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Applies `format` method of all `formaters` sequentially starting from the first
        formater in the list.
        """
        return functools.reduce(lambda df, formatter: formatter.format(df), self.formatters, df)

    def _flatten_formatters_list(self) -> List[PandasFormatter]:
        flattened_formatters = []
        for formatter in self.formatters:
            if isinstance(formatter, CompositeFormatter):
                flattened_formatters += formatter._flatten_formatters_list()
            else:
                flattened_formatters += [formatter]
        return flattened_formatters

    def flatten(self) -> CompositeFormatter:
        return CompositeFormatter(formatters=self._flatten_formatters_list(), logger=self.logger)

    @use_loggers("logger")
    def log(self, logger: Logger, params_not_to_log: List[str] = ["logger"]) -> None:
        super().log.__wrapped__(self, logger, params_not_to_log=["logger", "formatters"])

        for formatter in self.formatters:
            formatter.log()


@dataclass
class PandasIdentityFormatter(PandasFormatter):
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def format(self, df: pd.DataFrame) -> pd.DataFrame:
        return df


@dataclass
class PandasFloatCaster(PandasFormatter):
    target_dtype: str
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def format(self, df: pd.DataFrame) -> pd.DataFrame:
        return cast_float_cols(df, self.target_dtype)


@dataclass
class PandasIntegerCaster(PandasFormatter):
    target_dtype: str
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def format(self, df: pd.DataFrame) -> pd.DataFrame:
        return cast_integer_cols(df, self.target_dtype)


@dataclass
class PandasColumnsSelector(PandasFormatter):
    relevant_columns: List[str]
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def format(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[self.relevant_columns]


@dataclass
class PandasNullDropper(PandasFormatter):
    non_nullable_columns: List[str] = field(default_factory=list)
    nullable_columns: List[str] = field(default_factory=list)
    ignore_index: bool = True
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def __post_init__(self):
        super().__post_init__()
        if self.non_nullable_columns and self.nullable_columns:
            raise ValueError("non_nullable_columns and nullable_columns are both non-empty")

    def format(self, df: pd.DataFrame) -> pd.DataFrame:
        columns_subset = (
            self.non_nullable_columns
            if self.non_nullable_columns
            else (set(df.columns) - set(self.nullable_columns))
        )
        return df.dropna(subset=columns_subset, ignore_index=self.ignore_index)


@dataclass
class PandasValuesReplacer(PandasFormatter):
    """Applies `pandas.DataFrame.replace` with corresponding attributes"""

    to_replace: Any
    value: Any
    logger: Logger = field(default=VoidLogger(), repr=False, compare=False)

    def format(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.replace(to_replace=self.to_replace, value=self.value)
